from .flux_selectors import HardFluxSelection, SoftFluxSelection
from .generic_selectors import (BernoulliSelection, BoxSelection, LowerBound,
                                SoftSelection, UnitySelection, UpperBound)
from .selection_probability import DummySelection, SelectionProbabilty
from .spatial_selection import SpatialSelection

__all__ = ["HardFluxSelection", "SoftFluxSelection", "BernoulliSelection", "LowerBound", "UpperBound",
                                "SoftSelection", "UnitySelection", "SelectionProbabilty", "SpatialSelection", "BoxSelection", "DummySelection"]



from typing import Dict, Optional, Union

from popsynth.utils.logging import setup_logger

log = setup_logger(__name__)


class Parameter(object):
    def __init__(
        self,
        default: Optional[float] = None,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        free: bool = True
    ):

        self.name = None  # type: Union[None, str]
        self._vmin = vmin  # type: Optional[float]
        self._vmax = vmax  # type: Optional[float]
        self._default = default  # type: Optional[float]
        self._free = free

    @property
    def default(self) -> Optional[float]:
        return self._default

    def __get__(self, obj, type=None) -> object:

        try:

            if obj._parameter_storage[self.name] is None:

                obj._parameter_storage[self.name] = self._default

            assert obj._parameter_storage[self.name] is not None, "parameters must have values!"

            return obj._parameter_storage[self.name]

        except(KeyError):

            obj._parameter_storage[self.name] = self._default

        assert obj._parameter_storage[self.name] is not None, "parameters must have values!"

        return obj._parameter_storage[self.name]

    def __set__(self, obj, value) -> None:
        self._is_set = True

        if not self._free:

            log.error(f"{self.name} is fixed and cannot me set")
            raise RuntimeError()

        if self._vmin is not None:
            if not (value >= self._vmin):
                log.error(
                    f"trying to set {self.x} to a value below {self._vmin} is not allowed")

                raise RuntimeError()

        if self._vmax is not None:
            if not (value <= self._vmax):

                log.error(
                    f"trying to set {self.name} to a value above {self._vmax} is not allowed")
                raise RuntimeError()

        obj._parameter_storage[self.name] = value

        # Define property "free"

    def _set_free(self, value=True):

        self._free = value

    def _get_free(self):

        return self._free

    free = property(
        _get_free,
        _set_free,
        doc="Gets or sets whether the parameter is free or not. Use booleans, like: 'p.free = True' "
        " or 'p.free = False'. ",
    )

    # Define property "fix"

    def _set_fix(self, value=True):

        self._free = not value

    def _get_fix(self):

        return not self._free

    fix = property(
        _get_fix,
        _set_fix,
        doc="Gets or sets whether the parameter is fixed or not. Use booleans, like: 'p.fix = True' "
        " or 'p.fix = False'. ",
    )


class ParameterMeta(type):
    def __new__(mcls, name, bases, attrs, **kwargs):

        cls = super().__new__(mcls, name, bases, attrs, **kwargs)

        # Compute set of abstract method names
        abstracts = {
            name
            for name, value in attrs.items()
            if getattr(value, "__isabstractmethod__", False)
        }
        for base in bases:
            for name in getattr(base, "__abstractmethods__", set()):
                value = getattr(cls, name, None)
                if getattr(value, "__isabstractmethod__", False):
                    abstracts.add(name)
        cls.__abstractmethods__ = frozenset(abstracts)

        # parameters

        cls._parameter_storage = {}

        for k, v in attrs.items():

            if isinstance(v, Parameter):
                v.name = k

        return cls

    def __subclasscheck__(cls, subclass):
        """Override for issubclass(subclass, cls)."""
        if not isinstance(subclass, type):
            raise TypeError("issubclass() arg 1 must be a class")
        # Check cache

        # Check the subclass hook
        ok = cls.__subclasshook__(subclass)
        if ok is not NotImplemented:
            assert isinstance(ok, bool)
            if ok:
                cls._abc_cache.add(subclass)
            else:
                cls._abc_negative_cache.add(subclass)
            return ok
        # Check if it's a direct subclass
        if cls in getattr(subclass, "__mro__", ()):
            cls._abc_cache.add(subclass)
            return True
        # Check if it's a subclass of a registered class (recursive)
        for rcls in cls._abc_registry:
            if issubclass(subclass, rcls):
                cls._abc_cache.add(subclass)
                return True
        # Check if it's a subclass of a subclass (recursive)
        for scls in cls.__subclasses__():
            if issubclass(subclass, scls):
                cls._abc_cache.add(subclass)
                return True
        # No dice; update negative cache
        cls._abc_negative_cache.add(subclass)
        return False



from tqdm.auto import tqdm

from popsynth.utils.configuration import popsynth_config

_purple = "#B833FF"


def progress_bar(itr, **kwargs):

    return (tqdm(itr, colour=_purple, **kwargs) if popsynth_config.show_progress else itr)



from popsynth.populations.spatial_populations import (
    SphericalPopulation,
    ZPowerSphericalPopulation,
    ZPowerCosmoPopulation,
    SFRPopulation,
)

from popsynth.distributions.schechter_distribution import SchechterDistribution


class SchechterHomogeneousSphericalPopulation(SphericalPopulation):
    def __init__(self, Lambda, Lmin, alpha, r_max=10, seed=1234):
        """FIXME! briefly describe function

        :param Lambda:
        :param Lmin:
        :param alpha:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = SchechterDistribution(seed=seed)
        luminosity_distribution.Lmin = Lmin
        luminosity_distribution.alpha = alpha

        super(SchechterHomogeneousSphericalPopulation, self).__init__(
            Lambda=Lambda,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
        )


class SchechterZPowerSphericalPopulation(ZPowerSphericalPopulation):
    def __init__(self, Lambda, delta, Lmin, alpha, r_max=10, seed=1234):
        """FIXME! briefly describe function

        :param Lambda:
        :param delta:
        :param Lmin:
        :param alpha:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = SchechterDistribution(seed=seed)

        luminosity_distribution.Lmin = Lmin
        luminosity_distribution.alpha = alpha

        super(SchechterZPowerSphericalPopulation, self).__init__(
            Lambda=Lambda,
            delta=delta,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
        )


class SchechterZPowerCosmoPopulation(ZPowerCosmoPopulation):
    def __init__(
        self,
        Lambda,
        delta,
        Lmin,
        alpha,
        r_max=10,
        seed=1234,
        is_rate=True,
    ):
        """FIXME! briefly describe function

        :param Lambda:
        :param delta:
        :param Lmin:
        :param alpha:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = SchechterDistribution(seed=seed)

        luminosity_distribution.Lmin = Lmin
        luminosity_distribution.alpha = alpha

        super(SchechterZPowerCosmoPopulation, self).__init__(
            Lambda=Lambda,
            delta=delta,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
            is_rate=is_rate,
        )


class SchechterSFRPopulation(SFRPopulation):
    def __init__(
        self,
        r0,
        rise,
        decay,
        peak,
        Lmin,
        alpha,
        r_max=10,
        seed=1234,
        is_rate=True,
    ):
        """FIXME! briefly describe function

        :param r0:
        :param rise:
        :param decay:
        :param peak:
        :param Lmin:
        :param alpha:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = SchechterDistribution(seed=seed)

        luminosity_distribution.Lmin = Lmin
        luminosity_distribution.alpha = alpha

        super(SchechterSFRPopulation, self).__init__(
            r0=r0,
            rise=rise,
            decay=decay,
            peak=peak,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
            is_rate=is_rate,
        )



from .viewing_angle_sampler import ViewingAngleSampler
from .lognormal_aux_sampler import LogNormalAuxSampler, Log10NormalAuxSampler
from .normal_aux_sampler import NormalAuxSampler
from .trunc_normal_aux_sampler import TruncatedNormalAuxSampler

__all__ = ["ViewingAngleSampler",
           "LogNormalAuxSampler",
           "Log10NormalAuxSampler",
           "NormalAuxSampler",
           "TruncatedNormalAuxSampler"
           
           ]



from popsynth.populations.spatial_populations import (
    SphericalPopulation,
    ZPowerSphericalPopulation,
    ZPowerCosmoPopulation,
    SFRPopulation,
)
from popsynth.distributions.bpl_distribution import BPLDistribution


class BPLHomogeneousSphericalPopulation(SphericalPopulation):
    def __init__(self,
                 Lambda,
                 Lmin,
                 alpha,
                 Lbreak,
                 beta,
                 Lmax,
                 r_max=5,
                 seed=1234):
        """FIXME! briefly describe function

        :param Lambda:
        :param Lmin:
        :param alpha:
        :param Lbreak:
        :param beta:
        :param Lmax:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """
        luminosity_distribution = BPLDistribution(seed=seed)
        luminosity_distribution.Lmin = Lmin
        luminosity_distribution.alpha = alpha
        luminosity_distribution.Lbreak = Lbreak
        luminosity_distribution.beta = beta
        luminosity_distribution.Lmax = Lmax

        super(BPLHomogeneousSphericalPopulation, self).__init__(
            Lambda=Lambda,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
        )


class BPLZPowerSphericalPopulation(ZPowerSphericalPopulation):
    def __init__(self,
                 Lambda,
                 delta,
                 Lmin,
                 alpha,
                 Lbreak,
                 beta,
                 Lmax,
                 r_max=5,
                 seed=1234):
        """FIXME! briefly describe function

        :param Lambda:
        :param delta:
        :param Lmin:
        :param alpha:
        :param Lbreak:
        :param beta:
        :param Lmax:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """
        luminosity_distribution = BPLDistribution(seed=seed)
        luminosity_distribution.Lmin = Lmin
        luminosity_distribution.alpha = alpha
        luminosity_distribution.Lbreak = Lbreak
        luminosity_distribution.beta = beta
        luminosity_distribution.Lmax = Lmax

        super(BPLZPowerSphericalPopulation, self).__init__(
            Lambda=Lambda,
            delta=delta,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
        )


class BPLZPowerCosmoPopulation(ZPowerCosmoPopulation):
    def __init__(
        self,
        Lambda,
        delta,
        Lmin,
        alpha,
        Lbreak,
        beta,
        Lmax,
        r_max=5,
        seed=1234,
        is_rate=True,
    ):
        """FIXME! briefly describe function

        :param Lambda:
        :param delta:
        :param Lmin:
        :param alpha:
        :param Lbreak:
        :param beta:
        :param Lmax:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = BPLDistribution(seed=seed)
        luminosity_distribution.Lmin = Lmin
        luminosity_distribution.alpha = alpha
        luminosity_distribution.Lbreak = Lbreak
        luminosity_distribution.beta = beta
        luminosity_distribution.Lmax = Lmax

        super(BPLZPowerCosmoPopulation, self).__init__(
            Lambda=Lambda,
            delta=delta,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
            is_rate=is_rate,
        )


class BPLSFRPopulation(SFRPopulation):
    def __init__(
        self,
        r0,
        rise,
        decay,
        peak,
        Lmin,
        alpha,
        Lbreak,
        beta,
        Lmax,
        r_max=5,
        seed=1234,
        is_rate=True,
    ):
        """FIXME! briefly describe function

        :param r0:
        :param rise:
        :param decay:
        :param peak:
        :param Lmin:
        :param alpha:
        :param Lbreak:
        :param beta:
        :param Lmax:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = BPLDistribution(seed=seed)
        luminosity_distribution.Lmin = Lmin
        luminosity_distribution.alpha = alpha
        luminosity_distribution.Lbreak = Lbreak
        luminosity_distribution.beta = beta
        luminosity_distribution.Lmax = Lmax

        super(BPLSFRPopulation, self).__init__(
            r0=r0,
            rise=rise,
            decay=decay,
            peak=peak,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
            is_rate=is_rate,
        )



from popsynth.distribution import SpatialDistribution

from .selection_probability import SelectionProbabilty


class SpatialSelection(SelectionProbabilty):

    _selection_name = "SpatialSelection"

    def __init__(self, name) -> None:

        super(SpatialSelection, self).__init__(name)

        self._spatial_distribution: SpatialDistribution = None

    def set_spatial_distribution(
            self, spatial_distribtuion: SpatialDistribution) -> None:
        """
        set the spatial distribution
        """
        assert isinstance(spatial_distribtuion, SpatialDistribution)

        self._spatial_distribution: SpatialDistribution = spatial_distribtuion


__all__ = ["SpatialSelection"]



import scipy.stats as stats
import numpy as np

from popsynth.auxiliary_sampler import AuxiliarySampler, AuxiliaryParameter


class TruncatedNormalAuxSampler(AuxiliarySampler):
    _auxiliary_sampler_name = "TruncatedNormalAuxSampler"

    mu = AuxiliaryParameter(default=0)
    tau = AuxiliaryParameter(default=1, vmin=0)
    lower = AuxiliaryParameter()
    upper = AuxiliaryParameter()
    sigma = AuxiliaryParameter(default=1, vmin=0)

    def __init__(self, name, observed=True):
        """FIXME! briefly describe function

        :param name:
        :param observed:
        :returns:
        :rtype:

        """

        super(TruncatedNormalAuxSampler, self).__init__(name=name,
                                                        observed=observed)

    def true_sampler(self, size):

        l = (self.lower - self.mu) / self.tau
        u = (self.upper - self.mu) / self.tau

        self._true_values = stats.truncnorm.rvs(
            l,
            u,
            loc=self.mu,
            scale=self.tau,
            size=size,
        )

        assert np.alltrue(self._true_values >= self.lower)
        assert np.alltrue(self._true_values <= self.upper)

    def observation_sampler(self, size):

        if self._is_observed:

            self._obs_values = stats.norm.rvs(loc=self._true_values,
                                              scale=self.sigma,
                                              size=size)

        else:

            self._obs_values = self._true_values



import numpy as np
from scipy import random as rd

from popsynth.distribution import DistributionParameter
from popsynth.distributions.spherical_distribution import SphericalDistribution


class SpiralGalaxyDistribution(SphericalDistribution):
    _distribution_name = "SpiralGalaxyDistribution"

    rho = DistributionParameter(vmin=0)
    a = DistributionParameter()
    b = DistributionParameter()
    R1 = DistributionParameter()
    R0 = DistributionParameter()

    def __init__(self, seed=1234, name="spiral_galaxy", form=None):

        super(SpiralGalaxyDistribution, self).__init__(seed=seed,
                                                       name=name,
                                                       form=form)

    def dNdV(self, r):

        return (self.rho * np.power(
            (r + self.R1) / (self.R0 + self.R1),
            self.a,
        ) * np.exp(-self.b * (r - self.R0) / (self.R0 + self.R1)))

    def draw_sky_positions(self, size):

        # code thanks to Moritz!

        k = np.array([4.25, 4.25, 4.89, 4.89
                      ])  # Parameter fuer possun [-8.5,0,0], von Wainscoat1992
        r0 = np.array([3.48, 3.48, 4.90, 4.90])
        theta0 = np.array([0.0, 3.141, 2.525, 5.666])  # Wainscoat 1992
        height = 0.3  # kpc

        idx = np.random.randint(4, size=size)

        ks = k[idx]
        r0s = r0[idx]
        theta0s = theta0[idx]

        tet = ks * np.log(self._distances / r0s) + theta0s

        winklcor = np.random.uniform(0.0, 2 * np.pi, size=size)

        corrtet = winklcor * np.exp(
            -0.35 * self._distances)  # Faucher-Giguere 2007

        spiraltheta = tet + corrtet  # Faucher-Giguere 2007

        spiralthetaindec = spiraltheta * 180 / np.pi

        zpos = rd.exponential(height, size=size)

        zpos *= np.random.choice([-1, 1], size=size)

        self._distances = self._distances + np.random.normal(
            0, scale=0.07 * np.abs(self._distances), size=size)

        phi = np.arccos(zpos / np.sqrt(self._distances**2 + zpos**2))

        self._theta = spiraltheta
        self._phi = phi



import os
import warnings
from pathlib import Path
from dataclasses import dataclass

from omegaconf import OmegaConf

_config_path = Path("~/.config/popsynth/").expanduser()

_config_name = Path("popsynth_config.yml")

_config_file = _config_path / _config_name


# Define structure with dataclasses
@dataclass
class LogConsole:

    on: bool = True
    level: str = "WARNING"


@dataclass
class LogFile:

    on: bool = True
    level: str = "WARNING"


@dataclass
class Logging:

    debug: bool = False
    console: LogConsole = LogConsole()
    file: LogFile = LogFile()


@dataclass
class Cosmology:

    Om: float = 0.307
    h0: float = 67.7


@dataclass
class PopSynthConfig:

    logging: Logging = Logging()
    cosmology: Cosmology = Cosmology()
    show_progress: bool = True


# Read the default config
popsynth_config: PopSynthConfig = OmegaConf.structured(PopSynthConfig)

# Merge with local config
if _config_file.is_file():

    _local_config = OmegaConf.load(_config_file)

    popsynth_config: PopSynthConfig = OmegaConf.merge(popsynth_config,
                                                      _local_config)

# Write defaults
else:

    # Make directory if needed
    _config_path.mkdir(parents=True, exist_ok=True)

    with _config_file.open("w") as f:

        OmegaConf.save(config=popsynth_config, f=f.name)



#from numpy.typing import ArrayLike
from typing import List

import numpy as np
import scipy.stats as stats

from popsynth.auxiliary_sampler import AuxiliaryParameter, AuxiliarySampler

ArrayLike = List[float]


class LogNormalAuxSampler(AuxiliarySampler):
    _auxiliary_sampler_name = "LogNormalAuxSampler"

    mu = AuxiliaryParameter(default=0)
    tau = AuxiliaryParameter(default=1, vmin=0)
    sigma = AuxiliaryParameter(default=1, vmin=0)

    def __init__(self, name: str, observed: bool = True):
        """
        A Log normal sampler

        where x ~ 10^N(mu, sigma)


        :param name:
        :param observed:
        :returns:
        :rtype:

        """
        super(LogNormalAuxSampler, self).__init__(name=name, observed=observed)

    def true_sampler(self, size: int):

        self._true_values = np.exp(
            stats.norm.rvs(loc=self.mu, scale=self.tau,
                           size=size))  # type: ArrayLike

    def observation_sampler(self, size: int):

        if self._is_observed:

            self._obs_values = stats.norm.rvs(loc=self._true_values,
                                              scale=self.sigma,
                                              size=size)  # type: ArrayLike

        else:

            self._obs_values = self._true_values  # type: ArrayLike


class Log10NormalAuxSampler(AuxiliarySampler):
    _auxiliary_sampler_name = "Log10NormalAuxSampler"

    mu = AuxiliaryParameter(default=0)
    tau = AuxiliaryParameter(default=1, vmin=0)
    sigma = AuxiliaryParameter(default=1, vmin=0)

    def __init__(self, name: str, observed: bool = True):
        """
        A Log10 normal sampler.

        where x ~ 10^N(mu, sigma)



        :param name:
        :param observed:
        :returns:
        :rtype:

        """
        super(Log10NormalAuxSampler, self).__init__(name=name,
                                                    observed=observed)

    def true_sampler(self, size: int):

        self._true_values = np.power(10,
                                     stats.norm.rvs(
                                         loc=self.mu,
                                         scale=self.tau,
                                         size=size))  # type: ArrayLike

    def observation_sampler(self, size: int):

        if self._is_observed:

            self._obs_values = stats.norm.rvs(loc=self._true_values,
                                              scale=self.sigma,
                                              size=size)  # type: ArrayLike

        else:

            self._obs_values = self._true_values  # type: ArrayLike



import numpy as np


def sample_theta_phi(size: int):

    theta = np.arccos(1 - 2 * np.random.uniform(0.0, 1.0, size=size))
    phi = np.random.uniform(0, 2 * np.pi, size=size)

    return theta, phi


def xyz(r, theta, phi):
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)

    return x, y, z



import numpy as np

from popsynth.distribution import SpatialDistribution, DistributionParameter


class SphericalDistribution(SpatialDistribution):
    _distribution_name = "SphericalDistribution"
    def __init__(self, seed=1234, name="sphere", form=None):

        super(SphericalDistribution, self).__init__(seed=seed,
                                                    name=name,
                                                    form=form)

    def differential_volume(self, r):

        return 4 * np.pi * r * r

    def transform(self, L, r):

        r = r + 1

        return L / (4.0 * np.pi * r * r)


class ConstantSphericalDistribution(SphericalDistribution):
    _distribution_name = "ConstantSphericalDistribution"

    Lambda = DistributionParameter(default=1, vmin=0)

    def __init__(
        self,
        seed=1234,
        name="cons_sphere",
        form=None,
    ):

        if form is None:
            form = r"\Lambda"

        super(ConstantSphericalDistribution, self).__init__(seed=seed,
                                                            name=name,
                                                            form=form)

    def dNdV(self, distance):

        return self.Lambda


class ZPowerSphericalDistribution(ConstantSphericalDistribution):
    _distribution_name = "ZPowerSphericalDistribution"

    delta = DistributionParameter(default=1)

    def __init__(self, seed=1234, name="zpow_sphere"):

        spatial_form = r"\Lambda (z+1)^{\delta}"

        super(ZPowerSphericalDistribution, self).__init__(seed,
                                                          name,
                                                          form=spatial_form)

    def dNdV(self, distance):

        return self.Lambda * np.power(distance + 1.0, self.delta)



from .spatial_populations import (
    SphericalPopulation,
    SFRPopulation,
    ZPowerSphericalPopulation,
    ZPowerCosmoPopulation,
)
from .pareto_populations import (
    ParetoHomogeneousSphericalPopulation,
    ParetoSFRPopulation,
    ParetoZPowerCosmoPopulation,
    ParetoZPowerSphericalPopulation,
)

from .bpl_population import (
    BPLHomogeneousSphericalPopulation,
    BPLSFRPopulation,
    BPLZPowerCosmoPopulation,
    BPLZPowerSphericalPopulation,
)

from .schechter_populations import (
    SchechterHomogeneousSphericalPopulation,
    SchechterSFRPopulation,
    SchechterZPowerSphericalPopulation,
    SchechterZPowerCosmoPopulation,
)

from .lognormal_population import (
    Log10NormalSFRPopulation,
    Log10NormalHomogeneousSphericalPopulation,
    Log10NormalZPowerSphericalPopulation,
    Log10NormalZPowerCosmoPopulation,
    LogNormalHomogeneousSphericalPopulation,
    LogNormalSFRPopulation,
    LogNormalZPowerSphericalPopulation,
    LogNormalZPowerCosmoPopulation,
)

__all__ = [
    "SphericalPopulation",
    "SFRPopulation",
    "ZPowerSphericalPopulation",
    "ZPowerCosmoPopulation",
    "ParetoHomogeneousSphericalPopulation",
    "ParetoSFRPopulation",
    "ParetoZPowerSphericalPopulation",
    "ParetoZPowerCosmoPopulation",
    "BPLHomogeneousSphericalPopulation",
    "BPLSFRPopulation",
    "BPLZPowerSphericalPopulation",
    "BPLZPowerCosmoPopulation",
    "SchechterHomogeneousSphericalPopulation",
    "SchechterSFRPopulation",
    "SchechterZPowerSphericalPopulation",
    "SchechterZPowerCosmoPopulation",
    "Log10NormalHomogeneousSphericalPopulation",
    "Log10NormalSFRPopulation",
    "Log10NormalZPowerSphericalPopulation",
    "Log10NormalZPowerCosmoPopulation",
    "LogNormalHomogeneousSphericalPopulation",
    "LogNormalSFRPopulation",
    "LogNormalZPowerSphericalPopulation",
    "LogNormalZPowerCosmoPopulation",
]



import abc
from typing import Any, Dict, List, Optional, Union

import numpy as np
from class_registry import AutoRegister

from popsynth.distribution import SpatialContainer
from popsynth.selection_probability import SelectionProbabilty, UnitySelection
from popsynth.utils.cosmology import cosmology
from popsynth.utils.logging import setup_logger
from popsynth.utils.meta import Parameter, ParameterMeta
from popsynth.utils.registry import auxiliary_parameter_registry

#from numpy.typing import ArrayLike

log = setup_logger(__name__)

ArrayLike = List[float]
SamplerDict = Dict[str, Dict[str, ArrayLike]]


class AuxiliaryParameter(Parameter):
    pass


class AuxiliarySampler(object, metaclass=AutoRegister(auxiliary_parameter_registry, base_type=ParameterMeta)):
    def __init__(self,
                 name: str,
                 observed: bool = True,
                 uses_distance: bool = False,
                 uses_luminosity: bool = False,
                 uses_sky_position: bool = False) -> None:

        self._parameter_storage = {}  # type: Dict[str, float]
        self._name = name  # type: str
        self._obs_name = "%s_obs" % name  # type: str

        self._obs_values = None  # type: ArrayLike
        self._true_values = None  # type: ArrayLike
        self._is_observed = observed  # type: bool
        self._secondary_samplers = {}  # type: SamplerDict
        self._is_secondary = False  # type: bool
        self._parent_names = []
        self._has_secondary = False  # type: bool
        self._is_sampled = False  # type: bool
        self._selector = UnitySelection()  # type: SelectionProbabilty
        self._uses_distance = uses_distance  # type: bool
        self._uses_luminosity = uses_luminosity  # type: bool
        self._uses_sky_position = uses_sky_position  # type: bool

    def set_luminosity(self, luminosity: ArrayLike) -> None:
        """FIXME! briefly describe function

        :param luminosity:
        :returns:
        :rtype:

        """

        self._luminosity = luminosity  # type:ArrayLike

    def set_spatial_values(self, value: SpatialContainer) -> None:
        """FIXME! briefly describe function

        :param distance:
        :returns:
        :rtype:

        """

        self._distance = value.distance  # type:ArrayLike
        self._theta = value.theta
        self._phi = value.phi
        self._ra = value.ra
        self._dec = value.dec
        self._spatial_values = value

    def set_selection_probability(self, selector: SelectionProbabilty) -> None:

        assert isinstance(
            selector, SelectionProbabilty
        ), "The selector is not a valid selection probability"

        self._selector = selector  # type: SelectionProbabilty

    def _apply_selection(self) -> None:
        """
        Default selection if none is specfied in child class
        """

        self._selector.draw(len(self._obs_values))

    def set_secondary_sampler(self, sampler) -> None:
        """
        Allows the setting of a secondary sampler from which to derive values
        """

        # make sure we set the sampler as a secondary
        # this causes it to throw a flag in the main
        # loop if we try to add it again

        sampler.make_secondary(self.name)
        # attach the sampler to this class

        self._secondary_samplers[sampler.name] = sampler
        self._has_secondary = True  # type: bool

    def draw(self, size: int = 1):
        """
        Draw the primary and secondary samplers. This is the main call.

        :param size: the number of samples to draw
        """
        # do not resample!
        if not self._is_sampled:

            log.info(f"Sampling: {self.name}")

            if self._has_secondary:

                log.info(f"{self.name} is sampling its secondary quantities")

            self._selector.set_distance(self._distance)

            try:

                self._selector.set_luminosity(self._luminosity)

            except(AttributeError):

                log.debug("tried to set luminosity, but could not")

                pass

            for k, v in self._secondary_samplers.items():

                if not v.is_secondary:

                    log.error("Tried to sample a non-secondary, this is a bug")

                    raise RuntimeError()

                # we do not allow for the secondary
                # quantities to derive a luminosity
                # as it should be the last thing dervied

                log.debug(f"{k} will have it spatial values set")

                v.set_spatial_values(self._spatial_values)

                v.draw(size=size)

            # Now, it is assumed that if this sampler depends on the previous samplers,
            # then those properties have been drawn

            self.true_sampler(size=size)

            if self._is_observed:

                log.debug(f"{self.name} is sampling the observed values")

                self.observation_sampler(size)

            else:

                self._obs_values = (self._true_values)  # type: ArrayLike

            self._selector.set_observed_value(self._obs_values)

            # check to make sure we sampled!
            assert (self.true_values is not None
                    and len(self.true_values) == size
                    ), f"{self.name} likely has a bad true_sampler function"
            assert (self.obs_values is not None
                    and len(self.obs_values) == size
                    ), f"{self.name} likely has a observation_sampler function"

            # now apply the selection to yourself
            # if there is nothing coded, it will be
            # list of all true

            self._is_sampled = True

            self._apply_selection()

    def reset(self):
        """
        reset all the selections
        """

        if self._is_sampled:

            log.info(f"Auxiliary sampler: {self.name} is being reset")

            self._is_sampled = False
            self._obs_values = None  # type: ArrayLike
            self._true_values = None  # type: ArrayLike

            self._selector.reset()

        else:

            log.debug(
                f"{self.name} is not reseting as it has not been sampled")

        for k, v in self._secondary_samplers.items():

            v.reset()

    def make_secondary(self, parent_name: str) -> None:

        self._is_secondary = True  # type: bool
        self._parent_names.append(parent_name)

    def get_secondary_properties(
        self,
        recursive_secondaries: Optional[Dict[str, ArrayLike]] = None,
        graph=None,
        primary=None,
        spatial_distribution=None,
    ) -> SamplerDict:
        """FIXME! briefly describe function

        :param recursive_secondaries:
        :returns:
        :rtype:

        """

        # if a holder was not passed, create one
        if recursive_secondaries is None:

            recursive_secondaries = {}  # type: SamplerDict

        # now collect each property. This should keep recursing
        if self._has_secondary:

            for k, v in self._secondary_samplers.items():

                if graph is not None:

                    graph.add_node(k, observed=False)
                    graph.add_edge(k, primary)

                    if v.observed:
                        graph.add_node(v.obs_name, observed=False)
                        graph.add_edge(k, v.obs_name)

                    if v.uses_distance:

                        self._graph.add_edge(spatial_distribution.name, k)

                recursive_secondaries = v.get_secondary_properties(
                    recursive_secondaries, graph, k,
                    spatial_distribution)  # type: SamplerDict

        # add our own on
        recursive_secondaries[self._name] = {
            "true_values": self._true_values,
            "obs_values": self._obs_values,
            "selection": self._selector,
        }

        return recursive_secondaries

    @property
    def secondary_samplers(self) -> SamplerDict:
        """
        Secondary samplers
        """

        return self._secondary_samplers

    @property
    def is_secondary(self) -> bool:

        return self._is_secondary

    @property
    def parents(self) -> List[str]:
        return self._parent_names

    @property
    def has_secondary(self) -> bool:

        return self._has_secondary

    @property
    def observed(self) -> bool:
        """"""
        return self._is_observed

    @property
    def name(self) -> str:
        return self._name

    @property
    def obs_name(self) -> str:
        return self._obs_name

    @property
    def true_values(self) -> np.ndarray:
        """
        The true values

        :returns:
        :rtype:

        """

        return self._true_values

    @property
    def obs_values(self) -> np.ndarray:
        """
        The observed values
        :returns:
        :rtype:

        """

        return self._obs_values

    @property
    def selection(self) -> np.ndarray:
        """
        The selection function

        :returns:
        :rtype: np.ndarray

        """

        return self._selector.selection

    @property
    def selector(self) -> SelectionProbabilty:
        return self._selector

    @property
    def truth(self) -> Dict[str, float]:

        out = {}

        for k, v in self._parameter_storage.items():

            if v is not None:

                out[k] = v

        return out

    @property
    def uses_distance(self) -> bool:
        return self._uses_distance

    @property
    def uses_sky_position(self) -> bool:
        return self._uses_sky_position

    @property
    def uses_luminosity(self) -> np.ndarray:
        return self._luminosity

    @property
    def luminosity_distance(self):
        """
        luminosity distance if needed
        """

        return cosmology.luminosity_distance(self._distance)

    @abc.abstractmethod
    def true_sampler(self, size: int = 1):

        pass

    def observation_sampler(self, size: int = 1) -> np.ndarray:

        return self._true_values


class NonObservedAuxSampler(AuxiliarySampler):
    def __init__(self,
                 name: str,
                 uses_distance: bool = False,
                 uses_luminosity: bool = False):

        super(NonObservedAuxSampler, self).__init__(
            name=name,
            observed=False,
            uses_distance=uses_distance,
            uses_luminosity=uses_luminosity,
        )


class DerivedLumAuxSampler(AuxiliarySampler):
    def __init__(self, name: str, uses_distance: bool = False):
        """FIXME! briefly describe function

        :param name:
        :param sigma:
        :param observed:
        :returns:
        :rtype:

        """

        super(DerivedLumAuxSampler, self).__init__(name,
                                                   observed=False,
                                                   uses_distance=uses_distance)

    @abc.abstractmethod
    def compute_luminosity(self):

        raise RuntimeError("Must be implemented in derived class")



import numpy as np
import scipy.stats as stats

from popsynth.distribution import LuminosityDistribution, DistributionParameter


class LogNormalDistribution(LuminosityDistribution):
    _distribution_name = "LogNormalDistribution"

    mu = DistributionParameter()
    tau = DistributionParameter(vmin=0)

    def __init__(self, seed=1234, name="lognorm"):

        lf_form = r"\frac{\alpha L_{\rm min}^{\alpha}}{L^{\alpha+1}}"

        super(LogNormalDistribution, self).__init__(name=name,
                                                    seed=seed,
                                                    form=lf_form)

    def phi(self, L):

        return (1.0 / (self.tau * L * np.sqrt(2 * np.pi))) * np.exp(-(
            (np.log(L) - self.mu)**2) / (2 * self.tau**2))

    def draw_luminosity(self, size=1):

        return np.random.lognormal(mean=self.mu, sigma=self.tau, size=size)



from popsynth.populations.spatial_populations import (
    SphericalPopulation,
    ZPowerSphericalPopulation,
    ZPowerCosmoPopulation,
    SFRPopulation,
)

from popsynth.distributions.log10_normal_distribution import Log10NormalDistribution
from popsynth.distributions.log_normal_distribution import LogNormalDistribution


class LogNormalHomogeneousSphericalPopulation(SphericalPopulation):
    def __init__(self, Lambda, mu, tau, r_max=5, seed=1234):
        """FIXME! briefly describe function

        :param Lambda:
        :param mu:
        :param tau:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = LogNormalDistribution(seed=seed)
        luminosity_distribution.mu = mu
        luminosity_distribution.tau = tau

        super(LogNormalHomogeneousSphericalPopulation, self).__init__(
            Lambda=Lambda,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
        )


class LogNormalZPowerSphericalPopulation(ZPowerSphericalPopulation):
    def __init__(self, Lambda, delta, mu, tau, r_max=5, seed=1234):
        """FIXME! briefly describe function

        :param Lambda:
        :param delta:
        :param mu:
        :param tau:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = LogNormalDistribution(seed=seed)
        luminosity_distribution.mu = mu
        luminosity_distribution.tau = tau

        super(LogNormalZPowerSphericalPopulation, self).__init__(
            Lambda=Lambda,
            delta=delta,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
        )


class LogNormalZPowerCosmoPopulation(ZPowerCosmoPopulation):
    def __init__(
        self,
        Lambda,
        delta,
        mu,
        tau,
        r_max=5,
        seed=1234,
        is_rate=True,
    ):
        """FIXME! briefly describe function

        :param Lambda:
        :param delta:
        :param mu:
        :param tau:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = LogNormalDistribution(seed=seed)
        luminosity_distribution.mu = mu
        luminosity_distribution.tau = tau

        super(LogNormalZPowerCosmoPopulation, self).__init__(
            Lambda=Lambda,
            delta=delta,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
            is_rate=is_rate,
        )


class LogNormalSFRPopulation(SFRPopulation):
    def __init__(self,
                 r0,
                 rise,
                 decay,
                 peak,
                 mu,
                 tau,
                 r_max=5,
                 seed=1234,
                 is_rate=True):
        """FIXME! briefly describe function

        :param r0:
        :param rise:
        :param decay:
        :param peak:
         :param mu:
        :param tau:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = LogNormalDistribution(seed=seed)
        luminosity_distribution.mu = mu
        luminosity_distribution.tau = tau

        super(LogNormalSFRPopulation, self).__init__(
            r0=r0,
            rise=rise,
            decay=decay,
            peak=peak,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
            is_rate=is_rate,
        )


class Log10NormalHomogeneousSphericalPopulation(SphericalPopulation):
    def __init__(self, Lambda, mu, tau, r_max=5, seed=1234):
        """FIXME! briefly describe function

        :param Lambda:
        :param mu:
        :param tau:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = Log10NormalDistribution(seed=seed)
        luminosity_distribution.mu = mu
        luminosity_distribution.tau = tau

        super(Log10NormalHomogeneousSphericalPopulation, self).__init__(
            Lambda=Lambda,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
        )


class Log10NormalZPowerSphericalPopulation(ZPowerSphericalPopulation):
    def __init__(self, Lambda, delta, mu, tau, r_max=5, seed=1234):
        """FIXME! briefly describe function

        :param Lambda:
        :param delta:
        :param mu:
        :param tau:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = Log10NormalDistribution(seed=seed)
        luminosity_distribution.mu = mu
        luminosity_distribution.tau = tau

        super(Log10NormalZPowerSphericalPopulation, self).__init__(
            Lambda=Lambda,
            delta=delta,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
        )


class Log10NormalZPowerCosmoPopulation(ZPowerCosmoPopulation):
    def __init__(
        self,
        Lambda,
        delta,
        mu,
        tau,
        r_max=5,
        seed=1234,
        is_rate=True,
    ):
        """FIXME! briefly describe function

        :param Lambda:
        :param delta:
        :param mu:
        :param tau:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = Log10NormalDistribution(seed=seed)
        luminosity_distribution.mu = mu
        luminosity_distribution.tau = tau

        super(Log10NormalZPowerCosmoPopulation, self).__init__(
            Lambda=Lambda,
            delta=delta,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
            is_rate=is_rate,
        )


class Log10NormalSFRPopulation(SFRPopulation):
    def __init__(
        self,
        r0,
        rise,
        decay,
        peak,
        mu,
        tau,
        r_max=5,
        seed=1234,
        is_rate=True,
    ):
        """FIXME! briefly describe function

        :param r0:
        :param rise:
        :param decay:
        :param peak:
        :param mu:
        :param tau:
        :param r_max:
        :param seed:
        :returns:
        :rtype:

        """

        luminosity_distribution = Log10NormalDistribution(seed=seed)
        luminosity_distribution.mu = mu
        luminosity_distribution.tau = tau

        super(Log10NormalSFRPopulation, self).__init__(
            r0=r0,
            rise=rise,
            decay=decay,
            peak=peak,
            r_max=r_max,
            seed=seed,
            luminosity_distribution=luminosity_distribution,
            is_rate=is_rate,
        )



from popsynth.distributions.spherical_distribution import (
    ConstantSphericalDistribution,
    ZPowerSphericalDistribution,
)
from popsynth.distributions.cosmological_distribution import (
    SFRDistribution,
    # MergerDistribution,
    ZPowerCosmoDistribution,
)

from popsynth.distributions.spiral_galaxy_distribution import SpiralGalaxyDistribution

from popsynth.population_synth import PopulationSynth

## First create a bunch of spatial distributions


class SphericalPopulation(PopulationSynth):
    def __init__(self,
                 Lambda,
                 r_max=5.0,
                 seed=1234,
                 luminosity_distribution=None):
        """FIXME! briefly describe function

        :param Lambda:
        :param r_max:
        :param seed:
        :param luminosity_distribution:
        :returns:
        :rtype:

        """

        spatial_distribution = ConstantSphericalDistribution(seed=seed)
        spatial_distribution.Lambda = Lambda
        spatial_distribution.r_max = r_max

        super(SphericalPopulation, self).__init__(
            spatial_distribution=spatial_distribution,
            luminosity_distribution=luminosity_distribution,
            seed=seed,
        )


class ZPowerSphericalPopulation(PopulationSynth):
    def __init__(self,
                 Lambda,
                 delta,
                 r_max=5.0,
                 seed=1234,
                 luminosity_distribution=None):
        """FIXME! briefly describe function

        :param Lambda:
        :param delta:
        :param r_max:
        :param seed:
        :param luminosity_distribution:
        :returns:
        :rtype:

        """

        spatial_distribution = ZPowerSphericalDistribution(seed=seed)

        spatial_distribution.Lambda = Lambda
        spatial_distribution.r_max = r_max
        spatial_distribution.delta = delta

        super(ZPowerSphericalPopulation, self).__init__(
            spatial_distribution=spatial_distribution,
            luminosity_distribution=luminosity_distribution,
            seed=seed,
        )


class ZPowerCosmoPopulation(PopulationSynth):
    def __init__(
        self,
        Lambda,
        delta,
        r_max=5.0,
        seed=1234,
        luminosity_distribution=None,
        is_rate=True,
    ):
        """FIXME! briefly describe function

        :param Lambda:
        :param delta:
        :param r_max:
        :param seed:
        :param luminosity_distribution:
        :returns:
        :rtype:

        """

        spatial_distribution = ZPowerCosmoDistribution(seed=seed,
                                                       is_rate=is_rate)

        spatial_distribution.Lambda = Lambda
        spatial_distribution.r_max = r_max
        spatial_distribution.delta = delta

        super(ZPowerCosmoPopulation, self).__init__(
            spatial_distribution=spatial_distribution,
            luminosity_distribution=luminosity_distribution,
            seed=seed,
        )


class SFRPopulation(PopulationSynth):
    def __init__(
        self,
        r0,
        rise,
        decay,
        peak,
        r_max=5,
        seed=1234,
        luminosity_distribution=None,
        is_rate=True,
    ):
        """FIXME! briefly describe function

        :param r0:
        :param rise:
        :param decay:
        :param peak:
        :param r_max:
        :param seed:
        :param luminosity_distribution:
        :returns:
        :rtype:

        """

        spatial_distribution = SFRDistribution(seed=seed, is_rate=is_rate)

        spatial_distribution.r0 = r0
        spatial_distribution.rise = rise
        spatial_distribution.decay = decay
        spatial_distribution.peak = peak
        spatial_distribution.r_max = r_max

        super(SFRPopulation, self).__init__(
            spatial_distribution=spatial_distribution,
            luminosity_distribution=luminosity_distribution,
            seed=seed,
        )


class MWRadialPopulation(PopulationSynth):
    def __init__(
        self,
        rho,
        a=1.64,
        b=4.01,
        R1=0.55,
        R0=8.5,
        r_max=20,
        seed=1234,
        luminosity_distribution=None,
    ):

        spatial_distribution = SpiralGalaxyDistribution(seed=seed)

        spatial_distribution.rho = rho
        spatial_distribution.a = a
        spatial_distribution.b = b
        spatial_distribution.R1 = R1
        spatial_distribution.R0 = R0
        spatial_distribution.r_max = r_max

        super(MWRadialPopulation, self).__init__(
            spatial_distribution,
            luminosity_distribution=luminosity_distribution,
            seed=seed,
        )



import numpy as np
import scipy.stats as stats
from popsynth.distribution import LuminosityDistribution, DistributionParameter


class BPLDistribution(LuminosityDistribution):
    _distribution_name = "BPLDistribution"

    Lmin = DistributionParameter(vmin=0)
    alpha = DistributionParameter()
    Lbreak = DistributionParameter(vmin=0)
    beta = DistributionParameter()
    Lmax = DistributionParameter(vmin=0)

    def __init__(self, seed=1234, name="bpl"):

        lf_form = r"\frac{\alpha L_{\rm min}^{\alpha}}{L^{\alpha+1}}"

        super(BPLDistribution, self).__init__(seed=seed,
                                              name=name,
                                              form=lf_form)

    def phi(self, L):

        return bpl(L, self.Lmin, self.Lbreak, self.Lmax, self.alpha, self.beta)

    def draw_luminosity(self, size=1):

        u = np.atleast_1d(np.random.uniform(size=size))

        return sample_bpl(u, self.Lmin, self.Lbreak, self.Lmax, self.alpha,
                          self.beta)


def integrate_pl(x0, x1, x2, a1, a2):
    """
    x0: lower bound
    x1: break point
    x2: upper bound
    a1: lower power law index
    a2: upper power low index

    """

    # compute the integral of each piece analytically
    int_1 = (np.power(x1, a1 + 1.0) - np.power(x0, a1 + 1.0)) / (a1 + 1)
    int_2 = (np.power(x1, a1 - a2) *
             (np.power(x2, a2 + 1.0) - np.power(x1, a2 + 1.0)) / (a2 + 1))

    # compute the total integral
    total = int_1 + int_2

    # compute the weights of each piece of the function
    w1 = int_1 / total
    w2 = int_2 / total

    return w1, w2, total


def bpl(x, x0, x1, x2, a1, a2):
    """
    x: the domain of the function
    x0: lower bound
    x1: break point
    x2: upper bound
    a1: lower power law index
    a2: upper power low index

    """

    # creatre a holder for the values
    out = np.empty_like(x)

    # get the total integral to compute the normalization
    _, _, C = integrate_pl(x0, x1, x2, a1, a2)
    norm = 1.0 / C

    # create an index to select each piece of the function
    idx = x < x1

    # compute the lower power law
    out[idx] = np.power(x[idx], a1)

    # compute the upper power law
    out[~idx] = np.power(x[~idx], a2) * np.power(x1, a1 - a2)

    return out * norm


def sample_bpl(u, x0, x1, x2, a1, a2):
    """
    u: uniform random number between on {0,1}
    x0: lower bound
    x1: break point
    x2: upper bound
    a1: lower power law index
    a2: upper power low index
    """

    # compute the weights with our integral function
    w1, w2, _ = integrate_pl(x0, x1, x2, a1, a2)

    # create a holder array for our output
    out = np.empty_like(u)

    # compute the bernoulli trials for lower piece of the function
    # *if we wanted to do the upper part... we just reverse our index*
    # We also compute these to bools for numpy array selection
    idx = stats.bernoulli.rvs(w1, size=len(u)).astype(bool)

    # inverse transform sample the lower part for the "successes"
    out[idx] = np.power(
        u[idx] * (np.power(x1, a1 + 1.0) - np.power(x0, a1 + 1.0)) +
        np.power(x0, a1 + 1.0),
        1.0 / (1 + a1),
    )

    # inverse transform sample the upper part for the "failures"
    out[~idx] = np.power(
        u[~idx] * (np.power(x2, a2 + 1.0) - np.power(x1, a2 + 1.0)) +
        np.power(x1, a2 + 1.0),
        1.0 / (1 + a2),
    )

    return out



from class_registry import AutoRegister, ClassRegistry

auxiliary_parameter_registry = ClassRegistry("_auxiliary_sampler_name")

distribution_registry = ClassRegistry("_distribution_name")

selection_registry = ClassRegistry("_selection_name")


def list_available_distributions():
    for k, v in distribution_registry.items():

        print(k)


def list_available_selection_functions():
    for k, v in selection_registry.items():

        print(k)


def list_available_auxiliary_samplers():
    for k, v in auxiliary_parameter_registry.items():

        print(k)


__all__ = ["auxiliary_parameter_registry",
           "distribution_registry", "distribution_registry"]



import numpy as np
import scipy.special as sf
import scipy.stats as stats

from popsynth.utils.configuration import popsynth_config
from popsynth.utils.logging import setup_logger
from popsynth.utils.progress_bar import progress_bar

from .selection_probability import SelectionParameter, SelectionProbabilty

log = setup_logger(__name__)


class UnitySelection(SelectionProbabilty):
    _selection_name = "UnitySelection"

    def __init__(self, name="unity"):
        """
        A selection that returns all unity
        """
        super(UnitySelection, self).__init__(name=name)

    def draw(self, size: int) -> None:

        self._selection = np.ones(size, dtype=int).astype(bool)


class BernoulliSelection(SelectionProbabilty):
    _selection_name = "BernoulliSelection"

    probability = SelectionParameter(vmin=0, vmax=1, default=0.5)

    def __init__(self) -> None:

        super(BernoulliSelection, self).__init__(name="Bernoulli")

    def draw(self, size: int) -> None:

        if popsynth_config.show_progress:

            self._selection = np.zeros(size, dtype=int).astype(
                bool)  # type: np.ndarray

            for i in progress_bar(range(size), desc=f"Selecting {self.name}"):

                # see if we detect the distance
                if stats.bernoulli.rvs(self.probability) == 1:

                    self._selection[i] = 1

        else:

            self._selection = stats.bernoulli.rvs(
                self._probability, size=size).astype(bool)  # type: np.ndarray


class BoxSelection(SelectionProbabilty):
    _selection_name = "BoxSelection"

    vmin = SelectionParameter()
    vmax = SelectionParameter()

    def __init__(self, name: str = "box selection", use_obs_value: bool = False, use_distance=False, use_luminosity=False, use_flux: bool = False):

        super(BoxSelection, self).__init__(name=name, use_distance=use_distance,
                                           use_luminosity=use_luminosity, use_obs_value=use_obs_value, use_flux=use_flux)

    def draw(self, size: int) -> np.ndarray:

        if self._use_distance:
            values = self._distance

        if self._use_obs_value:

            values = self._observed_value

        if self._use_luminosity:

            values = self._luminosity

        if self._use_flux:

            values = self._observed_flux

        self._selection = (values >= self.vmin) & (values <= self.vmax)


class LowerBound(SelectionProbabilty):
    _selection_name = "LowerBound"

    boundary = SelectionParameter()

    def __init__(self, name="Hard selection", use_obs_value: bool = False, use_distance=False, use_luminosity=False, use_flux: bool = False):
        """
        hard selection above the boundary
        """
        super(LowerBound, self).__init__(name=name, use_distance=use_distance,
                                         use_luminosity=use_luminosity, use_obs_value=use_obs_value, use_flux=use_flux
                                         )

    def draw(self, size: int) -> None:

        if self._use_distance:
            values = self._distance

        if self._use_obs_value:

            values = self._observed_value

        if self._use_luminosity:

            values = self._luminosity

        if self._use_flux:

            values = self._observed_flux

        self._selection = values >= self.boundary


class UpperBound(SelectionProbabilty):
    _selection_name = "UpperBound"

    boundary = SelectionParameter()

    def __init__(self, name="Hard selection", use_obs_value: bool = False, use_distance=False, use_luminosity=False, use_flux: bool = False):
        """
        hard selection below the boundary
        """
        super(UpperBound, self).__init__(name=name, use_distance=use_distance,
                                         use_luminosity=use_luminosity, use_obs_value=use_obs_value, use_flux=use_flux
                                         )

    def draw(self, size: int) -> None:

        if self._use_distance:
            values = self._distance

        if self._use_obs_value:

            values = self._observed_value

        if self._use_luminosity:

            values = self._luminosity

        if self._use_flux:

            values = self._observed_flux

        self._selection = values <= self.boundary


class SoftSelection(SelectionProbabilty):
    _selection_name = "SoftSelection"

    boundary = SelectionParameter()
    strength = SelectionParameter(vmin=0)

    def __init__(self, name="Soft Selection", use_obs_value: bool = False, use_distance=False, use_luminosity=False, use_flux: bool = False) -> None:
        """
        selection using an inverse logit function either on the 
        log are linear value of the parameter

        :param boundary: center of the logit
        :param strength: width of the logit
        """

        super(SoftSelection, self).__init__(name=name, use_distance=use_distance,
                                            use_luminosity=use_luminosity, use_obs_value=use_obs_value, use_flux=use_flux)

    def draw(self,
             size: int,
             use_log=True) -> None:

        if self._use_distance:
            values = self._distance

        if self._use_obs_value:

            values = self._observed_value

        if self._use_luminosity:

            values = self._luminosity

        if self._use_flux:

            values = self._observed_flux

        if not use_log:
            probs = sf.expit(self.strength *
                             (values - self.boundary))  # type: np.ndarray

        else:

            probs = sf.expit(self.strength *
                             (np.log10(values) -
                              np.log10(self.boundary)))  # type: np.ndarray

        self._selection = stats.bernoulli.rvs(
            probs, size=len(values)).astype(bool)

    __all__ = ["UnitySelection", "BernoulliSelection",
               "BoxSelection", "UpperBound", "UpperBound", "SoftSelection"]



import scipy.stats as stats

from popsynth.auxiliary_sampler import AuxiliarySampler, AuxiliaryParameter


class NormalAuxSampler(AuxiliarySampler):
    _auxiliary_sampler_name = "NormalAuxSampler"
    
    mu = AuxiliaryParameter(default=0)
    tau = AuxiliaryParameter(default=1, vmin=0)
    sigma = AuxiliaryParameter(default=1, vmin=0)

    def __init__(self, name: str, observed: bool = True):
        """FIXME! briefly describe function

        :param name:
        :param mu:
        :param tau:
        :param sigma:
        :param observed:
        :returns:
        :rtype:

        """

        super(NormalAuxSampler, self).__init__(name=name, observed=observed)

    def true_sampler(self, size: int):

        self._true_values = stats.norm.rvs(loc=self.mu,
                                           scale=self.tau,
                                           size=size)

    def observation_sampler(self, size: int):

        if self._is_observed:

            self._obs_values = stats.norm.rvs(loc=self._true_values,
                                              scale=self.sigma,
                                              size=size)

        else:

            self._obs_values = self._true_values



import logging
import logging.handlers as handlers
import sys
from typing import Dict, Optional
from contextlib import contextmanager
import colorama
from colorama import Back, Fore, Style

from popsynth.utils.configuration import popsynth_config
from popsynth.utils.package_data import get_path_of_log_dir, get_path_of_log_file

colorama.deinit()
colorama.init(strip=False)
## set up the console logging


class ColoredFormatter(logging.Formatter):
    """
    Colored log formatter.
    """
    def __init__(self,
                 *args,
                 colors: Optional[Dict[str, str]] = None,
                 **kwargs) -> None:
        """Initialize the formatter with specified format strings."""

        super().__init__(*args, **kwargs)

        self.colors = colors if colors else {}

    def format(self, record) -> str:
        """Format the specified record as text."""

        record.color = self.colors.get(record.levelname, "")
        record.reset = Style.RESET_ALL

        return super().format(record)


class MyFilter(object):
    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):
        return logRecord.levelno != self.__level


# now create the developer handler that rotates every day and keeps
# 10 days worth of backup
popsynth_dev_log_handler = handlers.TimedRotatingFileHandler(
    get_path_of_log_file("dev.log"), when="D", interval=1, backupCount=10)

# lots of info written out

_dev_formatter = logging.Formatter(
    "%(asctime)s | %(name)s | %(levelname)s| %(funcName)s | %(lineno)d | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

popsynth_dev_log_handler.setFormatter(_dev_formatter)
popsynth_dev_log_handler.setLevel(logging.DEBUG)

# now set up the usr log which will save the info

popsynth_usr_log_handler = handlers.TimedRotatingFileHandler(
    get_path_of_log_file("usr.log"), when="D", interval=1, backupCount=10)

popsynth_usr_log_handler.setLevel(popsynth_config["logging"]["file"]["level"])

# lots of info written out
_usr_formatter = logging.Formatter("%(asctime)s | %(levelname)s | %(message)s",
                                   datefmt="%Y-%m-%d %H:%M:%S")

popsynth_usr_log_handler.setFormatter(_usr_formatter)

# now set up the console logger

_console_formatter = ColoredFormatter(
    "{color} {levelname:8} {reset}| {color} {message} {reset}",
    style="{",
    datefmt="%Y-%m-%d %H:%M:%S",
    colors={
        "DEBUG": Fore.CYAN,
        "INFO": Fore.GREEN + Style.BRIGHT,
        "WARNING": Fore.YELLOW + Style.DIM,
        "ERROR": Fore.RED + Style.BRIGHT,
        "CRITICAL": Fore.RED + Back.WHITE + Style.BRIGHT,
    },
)

popsynth_console_log_handler = logging.StreamHandler(sys.stdout)
popsynth_console_log_handler.setFormatter(_console_formatter)
popsynth_console_log_handler.setLevel(
    popsynth_config["logging"]["console"]["level"])

warning_filter = MyFilter(logging.WARNING)



class LoggingState(object):

    def __init__(self, popsynth_usr_log_handler, popsynth_console_log_handler,
       
                 ):
        """
        A container to store the stat of the logs
        """

        # attach the log handlers

        self.popsynth_usr_log_handler = popsynth_usr_log_handler
        self.popsynth_console_log_handler = popsynth_console_log_handler


        # store their current states

        self.popsynth_usr_log_handler_state = popsynth_usr_log_handler.level
        self.popsynth_console_log_handler_state = popsynth_console_log_handler.level

    def _store_state(self):

        self.popsynth_usr_log_handler_state = popsynth_usr_log_handler.level
        self.popsynth_console_log_handler_state = popsynth_console_log_handler.level


    def restore_last_state(self):

        self.popsynth_usr_log_handler.setLevel(
            self.popsynth_usr_log_handler_state)
        self.popsynth_console_log_handler.setLevel(
            self.popsynth_console_log_handler_state)

    def silence_logs(self):

        # store the state
        self._store_state()

        # silence the logs

        self.popsynth_usr_log_handler.setLevel(
            logging.CRITICAL)
        self.popsynth_console_log_handler.setLevel(
            logging.CRITICAL)


    def loud_logs(self):

        # store the state
        self._store_state()

        # silence the logs

        self.popsynth_usr_log_handler.setLevel(
            logging.INFO)
        self.popsynth_console_log_handler.setLevel(
            logging.INFO)

    def debug_logs(self):

        # store the state
        self._store_state()

        # silence the logs
        self.popsynth_console_log_handler.setLevel(
            logging.DEBUG)



_log_state = LoggingState(popsynth_usr_log_handler, popsynth_console_log_handler,
                          )




def silence_warnings():
    """
    supress warning messages in console and file usr logs
    """

    popsynth_usr_log_handler.addFilter(warning_filter)
    popsynth_console_log_handler.addFilter(warning_filter)


def activate_warnings():
    """
    supress warning messages in console and file usr logs
    """

    popsynth_usr_log_handler.removeFilter(warning_filter)
    popsynth_console_log_handler.removeFilter(warning_filter)


def update_logging_level(level):

    popsynth_console_log_handler.setLevel(level)


def silence_logs():
    """
    Turn off all logging
    """

    # handle dev logs independently
    popsynth_dev_log_handler.setLevel(logging.CRITICAL)


    _log_state.silence_logs()


def show_progress_bars():

    popsynth_config.show_progress = True


def silence_progress_bars():

    popsynth_config.show_progress = False



    
def quiet_mode():
    """
    turn off all logging and progress bars
    """

    silence_progress_bars()

    # save state and silence
    silence_logs()


def loud_mode():
    """
    turn on all progress bars and logging
    """

    show_progress_bars()

    # save state and make loud
    _log_state.loud_logs()


def activate_logs():
    """
    re-activate silenced logs
    """

    # handle dev logs independently
    popsynth_dev_log_handler.setLevel(logging.DEBUG)


    _log_state.restore_last_state()


def debug_mode():
    """
    activate debug in the console
    """

    # store state and switch console to debug
    _log_state.debug_logs()

    


    
@contextmanager
def silence_console_log():

    current_console_logging_level = popsynth_console_log_handler.level
    current_usr_logging_level = popsynth_usr_log_handler.level

    popsynth_console_log_handler.setLevel(logging.ERROR)
    popsynth_usr_log_handler.setLevel(logging.ERROR)

    try:
        yield

    finally:

        popsynth_console_log_handler.setLevel(current_console_logging_level)
        popsynth_usr_log_handler.setLevel(current_usr_logging_level)



def setup_logger(name):

    # A logger with name name will be created
    # and then add it to the print stream
    log = logging.getLogger(name)

    # this must be set to allow debug messages through
    log.setLevel(logging.DEBUG)

    # add the handlers

    if popsynth_config["logging"]["debug"]:
        log.addHandler(popsynth_dev_log_handler)

    if popsynth_config["logging"]["console"]["on"]:

        log.addHandler(popsynth_console_log_handler)

    if popsynth_config["logging"]["file"]["on"]:
        log.addHandler(popsynth_usr_log_handler)

    # we do not want to duplicate teh messages in the parents
    log.propagate = False

    return log



import abc
from dataclasses import dataclass
from typing import Dict, List, Union

import numpy as np
from class_registry import AutoRegister

from popsynth.utils.configuration import popsynth_config
from popsynth.utils.logging import setup_logger
from popsynth.utils.meta import Parameter, ParameterMeta
from popsynth.utils.progress_bar import progress_bar
from popsynth.utils.registry import distribution_registry
from popsynth.utils.rejection_sample import rejection_sample
from popsynth.utils.spherical_geometry import sample_theta_phi

#from numpy.typing import ArrayLike

ArrayLike = List[float]

log = setup_logger(__name__)


class DistributionParameter(Parameter):
    pass


class Distribution(object, metaclass=AutoRegister(distribution_registry, base_type=ParameterMeta)):
    _distribution_name = "Distribution"

    def __init__(self, name: str, seed: int, form: str) -> None:
        """
        A distribution base class

        :param name: the name of the distribution
        :param seed: the random seed
        :param form: the latex form
        :param truth: dictionary holding true parameters
        :returns:
        :rtype:

        """

        self._parameter_storage: Dict[str, float] = {}

        self._seed = seed  # type: int
        self._name = name  # type: str
        self._form = form  # type: str

    @property
    def name(self) -> str:
        return self._name

    @property
    def form(self) -> str:
        return self._form

    @property
    def params(self) -> Dict[str, float]:
        return self._parameter_storage

    @property
    def truth(self) -> Dict[str, float]:

        out = {}

        for k, v in self._parameter_storage.items():

            if v is not None:

                out[k] = v

        return out


@dataclass
class SpatialContainer:

    distance: ArrayLike
    theta: ArrayLike
    phi: ArrayLike

    @property
    def dec(self) -> np.ndarray:
        return 90 - np.rad2deg(self.theta)

    @property
    def ra(self) -> np.ndarray:
        return np.rad2deg(self.phi)


class SpatialDistribution(Distribution):
    _distribution_name = "SpatialDistribution"

    r_max = DistributionParameter(vmin=0, default=10)

    def __init__(self, name: str, seed: int, form: Union[str, None] = None):
        """
        A spatial distribution such as a redshift
        distribution

        :param name: the name of the distribution
        :param r_max: the maximum distance to sample
        :param seed: the random seed
        :param form: the latex form

        """
        self._theta = None
        self._phi = None

        super(SpatialDistribution, self).__init__(name=name,
                                                  seed=seed,
                                                  form=form)

    @abc.abstractmethod
    def differential_volume(self, distance):
        """
        the differential volume

        :param distance:
        :returns:
        :rtype:

        """

        raise RuntimeError("Must be implemented in derived class")

    @abc.abstractmethod
    def dNdV(self, distance):

        raise RuntimeError("Must be implemented in derived class")
        pass

    def time_adjustment(self, r):
        """FIXME! briefly describe function

        :param r:
        :returns:
        :rtype:

        """

        return 1.0

    @abc.abstractmethod
    def transform(self, flux, distance):
        pass

    @property
    def theta(self) -> np.ndarray:
        return self._theta

    @property
    def phi(self) -> np.ndarray:
        return self._phi

    @property
    def dec(self) -> np.ndarray:
        return 90 - np.rad2deg(self._theta)

    @property
    def ra(self) -> np.ndarray:
        return np.rad2deg(self._phi)

    @property
    def distances(self) -> np.ndarray:
        return self._distances

    @property
    def spatial_values(self):

        return SpatialContainer(self._distances, self._theta, self._phi)

    def draw_sky_positions(self, size: int) -> None:

        self._theta, self._phi = sample_theta_phi(size)

    def draw_distance(self, size: int) -> None:
        """
        Draw the distances from the specified dN/dr model
        """

        # create a callback for the sampler
        dNdr = (lambda r: self.dNdV(r) * self.differential_volume(r) / self.
                time_adjustment(r))

        # find the maximum point
        tmp = np.linspace(0.0, self.r_max, 500,
                          dtype=np.float64)  # type: ArrayLike
        ymax = np.max(dNdr(tmp))  # type: float

        # rejection sampling the distribution
        r_out = []

        # slow

        if popsynth_config.show_progress:
            for i in progress_bar(range(size), desc="Drawing distances"):
                flag = True
                while flag:

                    # get am rvs from 0 to the max of the function

                    y = np.random.uniform(low=0, high=ymax)  # type: float

                    # get an rvs from 0 to the maximum distance

                    r = np.random.uniform(low=0,
                                          high=self.r_max)  # type: float

                    # compare them

                    if y < dNdr(r):
                        r_out.append(r)
                        flag = False
        else:

            r_out = rejection_sample(size, ymax, self.r_max,
                                     dNdr)  # type: ArrayLike

        self._distances = np.array(r_out)  # type: ArrayLike


class LuminosityDistribution(Distribution):
    _distribution_name = "LuminosityDistribtuion"

    def __init__(self, name: str, seed: int, form: Union[str, None] = None):
        """
        A luminosity distribution such as a
        distribution

        :param name: the name of the distribution
        :param seed: the random seed
        :param form: the latex form
        """

        super(LuminosityDistribution, self).__init__(
            name=name,
            seed=seed,
            form=form,
        )

    @abc.abstractmethod
    def phi(self, luminosity):
        """
        The functional form of the distribution

        :param L:
        :returns:
        :rtype:

        """

        raise RuntimeError("Must be implemented in derived class")

    @abc.abstractmethod
    def draw_luminosity(self, size):
        pass



# from numba import jit, njit, prange, float64
import numpy as np


# @jit(parallel=False, forceobj=True)
def rejection_sample(size, ymax, xmax, func):
    """FIXME! briefly describe function

    :param size:
    :param ymax:
    :param xmax:
    :param func:
    :returns:
    :rtype:

    """

    r_out = np.zeros(size)

    for i in range(size):

        while True:

            # get am rvs from 0 to the max of the function

            y = np.random.uniform(low=0, high=ymax)

            # get an rvs from 0 to the maximum distance

            r = np.random.uniform(low=0, high=xmax)

            # compare them

            if y < func(r):
                r_out[i] = r
                break

    return r_out



import numpy as np
import numba as nb
import math

from popsynth.distribution import SpatialDistribution, DistributionParameter
from popsynth.utils.cosmology import cosmology


class CosmologicalDistribution(SpatialDistribution):
    _distribution_name = "CosmologicalDistribution"
    def __init__(self,
                 seed=1234,
                 name="cosmo",
                 form=None,
                 truth={},
                 is_rate=True):

        super(CosmologicalDistribution, self).__init__(
            seed=seed,
            name=name,
            form=form,
        )
        self._is_rate = is_rate

    def differential_volume(self, z):

        return cosmology.differential_comoving_volume(z)

    def transform(self, L, z):

        return L / (4.0 * np.pi * cosmology.luminosity_distance(z)**2)

    def time_adjustment(self, z):
        if self._is_rate:
            return 1 + z
        else:
            return 1.0


class SFRDistribution(CosmologicalDistribution):
    _distribution_name = "SFRDistribution"

    r0 = DistributionParameter(vmin=0)
    rise = DistributionParameter()
    decay = DistributionParameter()
    peak = DistributionParameter(vmin=0)

    def __init__(self, seed=1234, name="sfr", is_rate=True):

        spatial_form = r"\rho_0 \frac{1+r \cdot z}{1+ \left(z/p\right)^d}"

        super(SFRDistribution, self).__init__(
            seed=seed,
            name=name,
            form=spatial_form,
            is_rate=is_rate,
        )

    def dNdV(self, z):
        return _dndv(
            z,
            self.r0,
            self.rise,
            self.decay,
            self.peak,
        )


@nb.njit(fastmath=True)
def _dndv(z, r0, rise, decay, peak):
    top = 1.0 + rise * z
    bottom = 1.0 + np.power(z / peak, decay)

    return r0 * top / bottom


class ZPowerCosmoDistribution(CosmologicalDistribution):
    _distribution_name = "ZPowerCosmoDistribution"

    Lambda = DistributionParameter(default=1, vmin=0)
    delta = DistributionParameter()

    def __init__(
        self,
        seed=1234,
        name="zpow_cosmo",
        is_rate=True,
    ):

        spatial_form = r"\Lambda (z+1)^{\delta}"

        super(ZPowerCosmoDistribution, self).__init__(
            seed=seed,
            name=name,
            form=spatial_form,
            is_rate=is_rate,
        )

    def dNdV(self, distance):

        return _zp_dndv(distance, self.Lambda, self.delta)

@nb.njit(fastmath=True)
def _zp_dndv(z, Lambda, delta):
    return Lambda * np.power(z + 1, delta)
    

    
# class MergerDistribution(CosmologicalDistribution):
#_distribution_name = "MergerDistribution"
#     def __init__(self, r0, td, sigma, r_max=10, seed=1234, name="merger"):

#         spatial_form = r"\rho_0 \frac{1+r \cdot z}{1+ \left(z/p\right)^d}"

#         super(MergerDistribution, self).__init__(
#             r_max=r_max, seed=seed, name=name, form=spatial_form
#         )

#         self._td = td
#         self._sigma = sigma

#     def _sfr(self, z):

#         top = 0.01 + 0.12 * z
#         bottom = 1.0 + np.power(z / 3.23, 4.66)
#         return top / bottom

#     def _delay_time(self, tau):
#         return np.exp(
#             -((np.log(tau) - np.log(self._td)) ** 2) / (2 * self._sigma ** 2)
#         ) / (np.sqrt(2 * np.pi) * self._sigma)

#     def dNdV(self, z):

#         integrand = lambda x: self._sfr(x) * self._delay_time(
#             cosmo.lookback_time(x).value
#         )
#         out = []

#         try:
#             return self.r0 * integrate.quad(integrand, z, np.infty)[0]

#         except:
#             for zz in z:

#                 out.append(integrate.quad(integrand, zz, np.infty)[0])

#             return self.r0 * np.array(out)



from pathlib import Path
import pkg_resources

from popsynth.utils.configuration import popsynth_config

def get_path_of_data_file(data_file) -> Path:
    file_path = pkg_resources.resource_filename("popsynth", "data/%s" % data_file)

    return Path(file_path)



def get_path_of_log_dir() -> Path:

    p: Path = Path("~/.log/popsynth").expanduser()

    if not p.exists():

        p.mkdir(parents=True)

    return p


def get_path_of_log_file(file_name: str) -> Path:

    return get_path_of_log_dir() / file_name



import numpy as np

from popsynth.auxiliary_sampler import (AuxiliaryParameter,
                                        NonObservedAuxSampler)


class ViewingAngleSampler(NonObservedAuxSampler):
    _auxiliary_sampler_name = "ViewingAngleSampler"

    max_angle = AuxiliaryParameter(default=90, vmin=0, vmax=180)

    def __init__(self):
        """
        A viewing angle sampler that samples from 0, max_angle.
        It assumes that this is NOT an observed property

        :param max_angle: the maximum angle to which to sample in DEGS
        :returns: None
        :rtype: None

        """

        super(ViewingAngleSampler, self).__init__(name="va", )

    def true_sampler(self, size: int) -> None:
        """
        Sample the viewing angle by inverse CDF

        :param size: number of samples
        :returns: None
        :rtype: None

        """

        theta_inverse = np.random.uniform(0.0,
                                          1 -
                                          np.cos(np.deg2rad(self.max_angle)),
                                          size=size)

        self._true_values = np.arccos(1.0 - theta_inverse)



import numpy as np

from popsynth.utils.logging import setup_logger

from .generic_selectors import LowerBound, SoftSelection

log = setup_logger(__name__)


class HardFluxSelection(LowerBound):
    _selection_name = "HardFluxSelection"

    def __init__(self) -> None:

        super(HardFluxSelection, self).__init__(use_flux=True)

    def draw(self, size: int):

        super(HardFluxSelection, self).draw(size)


class SoftFluxSelection(SoftSelection):
    _selection_name = "SoftFluxSelection"

    def __init__(self) -> None:

        super(SoftFluxSelection, self).__init__(use_flux=True)

    def draw(
        self,
        size: int,
    ):

        super(SoftFluxSelection, self).draw(size, use_log=True)


__all__ = ["HardFluxSelection", "SoftFluxSelection"]



import os

import h5py
import numpy as np


def recursively_save_dict_contents_to_group(h5file, path, dic):
    """
    load files from hdf5

    """
    for key, item in dic.items():
        if isinstance(
            item,
                (np.ndarray, np.int64, np.float64, str, bytes, float, int)):
            h5file[path + "/" + key] = item
        elif isinstance(item, dict):
            recursively_save_dict_contents_to_group(h5file,
                                                    path + "/" + key + "/",
                                                    item)
        else:
            raise ValueError("Cannot save %s type" % type(item))


def recursively_load_dict_contents_from_group(h5file, path):
    """
    load files from hdf5
    """
    ans = {}

    for key, item in h5file[path].items():

        if isinstance(item, h5py._hl.dataset.Dataset):
            ans[key] = item[()]
        elif isinstance(item, h5py._hl.group.Group):
            ans[key] = recursively_load_dict_contents_from_group(
                h5file, path + "/" + key + "/")
    return ans


# these two functions add and clean
# a networkx graph dict so that it will be stored in
# an hdf5 file


def fill_graph_dict(graph_dict):
    new_dict = {}

    for k, v in graph_dict.items():
        if len(v) == 0:
            # this is an empty dict
            # so fill it
            new_dict[k] = 1
        else:
            new_dict[k] = fill_graph_dict(v)
    return new_dict


def clean_graph_dict(graph_dict):
    new_dict = {}

    for k, v in graph_dict.items():
        if not isinstance(v, dict):
            # this is an empty dict
            # so fill it
            new_dict[k] = {}

        else:

            new_dict[k] = clean_graph_dict(v)
    return new_dict



import numpy as np

from popsynth.distribution import LuminosityDistribution, DistributionParameter


class ParetoDistribution(LuminosityDistribution):
    _distribution_name = "ParetoDistribution"

    Lmin = DistributionParameter(default=1, vmin=0)
    alpha = DistributionParameter(default=1)

    def __init__(self, seed=1234, name="pareto"):
        """
        A Pareto luminosity function
        :param seed:
        :param name:
        :returns:
        :rtype:

        """

        lf_form = r"\frac{\alpha L_{\rm min}^{\alpha}}{L^{\alpha+1}}"

        super(ParetoDistribution, self).__init__(seed=seed,
                                                 name=name,
                                                 form=lf_form)

    def phi(self, L):
        """
        The luminosity function

        :param L:
        :returns:
        :rtype:

        """

        out = np.zeros_like(L)

        idx = L >= self.Lmin

        out[idx] = self.alpha * self.Lmin**self.alpha / L[idx]**(self.alpha +
                                                                 1)

        return out

    def draw_luminosity(self, size=1):
        """FIXME! briefly describe function

        :param size:
        :returns:
        :rtype:

        """

        return (np.random.pareto(self.alpha, size) + 1) * self.Lmin



import numpy as np

from popsynth.distribution import SpatialDistribution, DistributionParameter


class FlatlandDistribution(SpatialDistribution):
    _distribution_name = "FlatlandDistribution"

    Lambda = DistributionParameter(default=1, vmin=0)

    def __init__(self, seed: int = 1234, name: str = "flatland", form=None):

        super(FlatlandDistribution, self).__init__(seed=seed,
                                                   name=name,
                                                   form=form)

    def differential_volume(self, r):

        return 1

    def transform(self, L, r):

        return L

    def dNdV(self, distance):

        return self.Lambda



from .bpl_distribution import BPLDistribution
from .cosmological_distribution import (CosmologicalDistribution,
                                        SFRDistribution,
                                        ZPowerCosmoDistribution)
from .log10_normal_distribution import Log10NormalDistribution
from .log_normal_distribution import LogNormalDistribution
from .pareto_distribution import ParetoDistribution
from .schechter_distribution import SchechterDistribution
from .spherical_distribution import (ConstantSphericalDistribution,
                                     SphericalDistribution,
                                     ZPowerSphericalDistribution)

__all__ = [
    "SphericalDistribution",
    "CosmologicalDistribution",
    "SFRDistribution",
    "ZPowerCosmoDistribution",
    "ParetoDistribution",
    "Log10NormalDistribution",
    "LogNormalDistribution",
    "SchechterDistribution",
    "BPLDistribution",
    "SphericalDistribution",
    "ConstantSphericalDistribution",
    "ZPowerSphericalDistribution"


]



import scipy.stats as stats

import popsynth.populations as populations
from popsynth.aux_samplers import *
from popsynth.auxiliary_sampler import (AuxiliaryParameter, AuxiliarySampler,
                                        DerivedLumAuxSampler,
                                        NonObservedAuxSampler)
from popsynth.distribution import (DistributionParameter,
                                   LuminosityDistribution, SpatialDistribution)
from popsynth.distributions import *
from popsynth.population import Population
from popsynth.population_synth import PopulationSynth
from popsynth.populations import *
from popsynth.selection_probability import *
from popsynth.utils.configuration import popsynth_config
from popsynth.utils.cosmology import cosmology
from popsynth.utils.logging import (activate_logs, activate_warnings,
                                    debug_mode, loud_mode, quiet_mode,
                                    show_progress_bars, silence_logs,
                                    silence_progress_bars, silence_warnings,
                                    update_logging_level)
from popsynth.utils.registry import (list_available_auxiliary_samplers,
                                     list_available_distributions,
                                     list_available_selection_functions)

__all__ = [
    "AuxiliarySampler",
    "DerivedLumAuxSampler",
    "populations",
    "Population",
    "PopulationSynth",
]

# chance = stats.bernoulli.rvs(0.3)

# if chance:

#     from IPython.display import display, YouTubeVideo

#     display(YouTubeVideo("FYJ1dbyDcrI"))

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



import scipy.special as sf
import numpy as np

from popsynth.distribution import LuminosityDistribution, DistributionParameter


class SchechterDistribution(LuminosityDistribution):
    _distribution_name = "SchechterDistribution"

    Lmin = DistributionParameter(default=1, vmin=0)
    alpha = DistributionParameter(default=1)

    def __init__(self, seed=1234, name="schechter"):
        """
        A Schechter luminosity function
        :param seed: the random number seed
        :param name: the name
        :returns: None
        :rtype:

        """

        form = r"\frac{1}{L_{\rm min}^{1+\alpha} \Gamma\left(1+\alpha\right)} L^{\alpha} \exp\left[ - \frac{L}{L_{\rm min}}\right]"

        super(SchechterDistribution, self).__init__(name=name,
                                                    seed=seed,
                                                    form=form)

    def phi(self, L):
        """FIXME! briefly describe function

        :param L:
        :returns:
        :rtype:

        """

        return (L**self.alpha * np.exp(-L / self.Lmin) /
                (self.Lmin**(1 + self.alpha) * sf.gamma(1 + self.alpha)))

    def draw_luminosity(self, size=1):
        """FIXME! briefly describe function

        :param size:
        :returns:
        :rtype:

        """

        xs = 1 - np.random.uniform(size=size)
        return self.Lmin * sf.gammaincinv(1 + self.alpha, xs)



import importlib
from typing import Any, Dict, List, Optional, Type

import h5py
import ipyvolume as ipv
import ipywidgets as widgets
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import pythreejs
from betagen import betagen
from IPython.display import Markdown, Math, display

from popsynth.utils.array_to_cmap import array_to_cmap
from popsynth.utils.hdf5_utils import (
    clean_graph_dict, fill_graph_dict,
    recursively_load_dict_contents_from_group,
    recursively_save_dict_contents_to_group)
from popsynth.utils.logging import setup_logger
from popsynth.utils.spherical_geometry import xyz

log = setup_logger(__name__)

# from numpy.typing import ArrayLike

# dummy
ArrayLike = List[float]

wine = "#8F2727"
dark, dark_highlight, mid, mid_highlight, light, light_highlight = betagen(
    wine)


class Population(object):
    def __init__(
        self,
        luminosities: ArrayLike,
        distances: ArrayLike,
        known_distances: ArrayLike,
        known_distance_idx: ArrayLike,
        unknown_distance_idx: ArrayLike,
        fluxes: ArrayLike,
        flux_obs: ArrayLike,
        selection: ArrayLike,
        flux_sigma: float,
        r_max: float,
        n_model: int,
        lf_params: Dict[str, Any],
        spatial_params: Optional[Dict[str, Any]] = None,
        model_spaces: Optional[ArrayLike] = None,
        seed: int = 1234,
        name: Optional[str] = None,
        spatial_form: Optional[Dict[str, Any]] = None,
        lf_form: Optional[Dict[str, Any]] = None,
        auxiliary_quantities: Optional[Dict[str, Any]] = None,
        truth: Dict[str, float] = {},
        graph: Optional[Dict[str, Any]] = None,
        theta=None,
        phi=None,
        pop_synth: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        A population containing all the simulated variables

        :param luminosities: the luminosities
        :param distances: the distances
        :param known_distances: the known distances
        :param known_distance_idx: the index of the known distances
        :param unknown_distance_idx:  the index of the unknown distances
        :param fluxes: the latent fluxes
        :param flux_obs: the observed fluxes
        :param selection: the selection vector
        :param flux_sigma: the uncertainty on the observed flux
        :param r_max: the maximum distance of the survey
        :param n_model:
        :param lf_params:
        :param spatial_params:
        :param model_spaces:
        :param seed: the random seed
        :param name:
        :param spatial_form:
        :param lf_form:
        :param auxiliary_quantities:
        :param truth:
        :returns:
        :rtype:

        """
        self._luminosities = luminosities  # type: ArrayLike

        self._distances = distances  # type: ArrayLike
        self._known_distances = known_distances  # type: ArrayLike
        self._known_distance_idx = known_distance_idx  # type: ArrayLike
        self._unknown_distance_idx = unknown_distance_idx  # type: ArrayLike

        self._theta = theta  # type: ArrayLike
        self._phi = phi  # type: ArrayLike

        assert len(known_distances) + len(unknown_distance_idx) == sum(
            selection), "the distances are not the correct size"

        # latent fluxes
        self._fluxes = fluxes  # type: ArrayLike

        # observed fluxes
        self._flux_obs = flux_obs  # type: ArrayLike
        self._selection = selection  # type: ArrayLike
        self._flux_sigma = flux_sigma  # type: float

        self._r_max = r_max  # type: float
        self._seed = seed  # type: int
        self._n_model = n_model  # type: int
        self._name = name  # type: str
        self._spatial_form = spatial_form
        self._lf_form = lf_form

        self._flux_selected = flux_obs[selection]  # type: ArrayLike
        self._distance_selected = distances[selection]  # type: ArrayLike
        self._luminosity_selected = luminosities[selection]  # type: ArrayLike

        self._flux_hidden = flux_obs[~selection]  # type: ArrayLike
        self._distance_hidden = distances[~selection]  # type: ArrayLike
        self._luminosity_hidden = luminosities[~selection]  # type: ArrayLike

        self._lf_params = lf_params
        self._spatial_params = spatial_params

        self._model_spaces = model_spaces

        self._truth = truth

        self._graph = graph

        self._n_objects = len(selection)  # type: int
        self._n_detections = sum(self._selection)  # type: int
        self._n_non_detections = self._n_objects - self._n_detections  # type: int

        self._pop_synth: Optional[Dict[str, Any]] = pop_synth

        if self._n_detections == 0:

            self._no_detection = True

            log.warning("THERE ARE NO DETECTED OBJECTS IN THE POPULATION")

        else:

            self._no_detection = False

        if auxiliary_quantities is not None:

            for k, v in auxiliary_quantities.items():

                setattr(self, k, v["true_values"])
                setattr(self, "%s_obs" % k, v["obs_values"])
                setattr(self, "%s_selected" % k, v["obs_values"][selection])

        self._auxiliary_quantities = auxiliary_quantities

        if model_spaces is not None:

            for k, v in model_spaces.items():

                assert len(v) == n_model

    @property
    def graph(self):
        return self._graph

    @property
    def pop_synth(self) -> Dict[str, Any]:
        """
        Dictionary population synth used to create this population
        """
        return self._pop_synth

    @property
    def truth(self):
        """
        the simulated truth parameters

        :returns:
        :rtype:

        """

        return self._truth

    @property
    def flux_sigma(self) -> float:

        return self._flux_sigma

    @property
    def theta(self) -> np.ndarray:
        return self._theta

    @property
    def phi(self) -> np.ndarray:
        return self._phi

    @property
    def dec(self) -> np.ndarray:
        return 90 - np.rad2deg(self._theta)

    @property
    def ra(self) -> np.ndarray:
        return np.rad2deg(self._phi)

    @property
    def luminosities_latent(self) -> np.ndarray:
        """
        The true luminosities of the objects. These are always latent
        as one cannot directly observe them
        """
        return self._luminosities

    @property
    def distances(self) -> np.ndarray:
        """
        The distances to the objects
        """
        return self._distances

    @property
    def known_distances(self) -> np.ndarray:
        """
        The observed distances
        """

        return self._known_distances

    @property
    def selection(self) -> np.ndarray:
        """
        The selection vector
        """
        return self._selection

    @property
    def fluxes_latent(self) -> np.ndarray:
        """
        The latent fluxes of the objects
        """
        return self._fluxes

    @property
    def fluxes_observed(self) -> np.ndarray:
        """
        All of the observed fluxes, i.e.,
        scattered with error

        """
        return self._flux_obs

    @property
    def selected_fluxes_observed(self) -> np.ndarray:
        """
        The selected obs fluxes
        """

        return self._flux_selected

    @property
    def selected_fluxes_latent(self) -> np.ndarray:
        """
        The selected latent fluxes
        """

        return self._fluxes[self._selection]

    @property
    def selected_distances(self) -> np.ndarray:
        """
        The selected distances. Note, this is different than
        the KNOWN distances
        """
        return self._distance_selected

    @property
    def hidden_fluxes_observed(self) -> np.ndarray:
        """
        The observed fluxes that are hidden by the selection
        """

        return self._flux_hidden

    @property
    def hidden_distances(self) -> np.ndarray:
        """
        The distances that are hidden by the selection
        """

        return self._distance_hidden

    @property
    def hidden_fluxes_latent(self) -> np.ndarray:
        """
        The latent fluxes that are hidden by the selection
        """

        return self._fluxes[~self._selection]

    @property
    def hard_cut(self) -> bool:
        return self._hard_cut

    @property
    def distance_probability(self) -> float:
        return self._distance_probability

    @property
    def luminosity_parameters(self):
        return self._lf_params

    @property
    def n_objects(self) -> int:
        """
        The number of objects in the population
        """
        return self._n_objects

    @property
    def n_detections(self) -> int:
        """
        The number of DETECTED objects in the population
        """

        return self._n_detections

    @property
    def n_non_detections(self) -> int:
        """
        The number of NON-DETECTED objects in the population
        """

        return self._n_non_detections

    @property
    def has_detections(self) -> bool:
        """
        If the population has detections
        """

        return not self._no_detection

    @property
    def spatial_parameters(self):
        return self._spatial_params

    def to_stan_data(self) -> dict:
        """
        Create Stan input
        """

        # create a dict for Stan
        output = dict(
            N=self._n_detections,
            Nz=len(self._known_distances),
            Nnz=len(self._unknown_distance_idx),
            z_obs=self._distance_selected,
            known_z_obs=self._known_distances,
            z_idx=self._known_distance_idx + 1,  # stan indexing
            z_nidx=self._unknown_distance_idx + 1,  # stan indexing
            r_obs=self._distance_selected,
            known_r_obs=self._known_distances,
            r_idx=self._known_distance_idx + 1,  # stan indexing
            r_nidx=self._unknown_distance_idx + 1,  # stan indexing
            log_flux_obs=np.log10(self._flux_selected),
            flux_obs=self._flux_selected,
            flux_sigma=self._flux_sigma,
            z_max=self._r_max,
            r_max=self._r_max,
            N_model=self._n_model,
        )

        # now append the model spaces
        for k, v in self._model_spaces.items():

            output[k] = v

        for k, v in self._auxiliary_quantities.items():

            output["%s_obs" % k] = v["obs_values"][self._selection]

        for k, v in output.items():

            if isinstance(v, np.int64):

                output[k] = int(v)

        return output

    def writeto(self, file_name: str) -> None:
        """
        write population to an HDF5 file

        :param file_name:
        :returns:
        :rtype:

        """

        with h5py.File(file_name, "w") as f:

            self._writeto(f)

    def addto(self, file_name: str, group_name: str) -> None:
        """
        write population to a group in an existing
        HDF5 file.
        """

        with h5py.File(file_name, "r+") as f:

            g = f.create_group(group_name)

            self._writeto(g)

    def _writeto(self, f):
        """
        write to file or group

        :param f: h5py file or group handle
        """

        spatial_grp = f.create_group("spatial_params")

        for k, v in self._spatial_params.items():

            spatial_grp.create_dataset(k,
                                       data=np.array([v]),
                                       compression="lzf")

        if self._lf_params is not None:

            lum_grp = f.create_group("lf_params")

            for k, v in self._lf_params.items():

                lum_grp.create_dataset(k,
                                       data=np.array([v]),
                                       compression="lzf")

            f.attrs["lf_form"] = np.string_(self._lf_form)

            f.attrs["has_lf"] = True

        else:
            f.attrs["has_lf"] = False

        f.attrs["name"] = np.string_(self._name)
        f.attrs["spatial_form"] = np.string_(self._spatial_form)
        f.attrs["flux_sigma"] = self._flux_sigma
        f.attrs["n_model"] = self._n_model
        f.attrs["r_max"] = self._r_max
        f.attrs["seed"] = int(self._seed)

        f.create_dataset("luminosities",
                         data=self._luminosities,
                         compression="lzf")
        f.create_dataset("distances", data=self._distances, compression="lzf")
        f.create_dataset("known_distances",
                         data=self._known_distances,
                         compression="lzf")
        f.create_dataset("known_distance_idx",
                         data=self._known_distance_idx,
                         compression="lzf")
        f.create_dataset(
            "unknown_distance_idx",
            data=self._unknown_distance_idx,
            compression="lzf",
        )
        f.create_dataset("fluxes", data=self._fluxes, compression="lzf")
        f.create_dataset("flux_obs", data=self._flux_obs, compression="lzf")
        f.create_dataset("selection", data=self._selection, compression="lzf")
        f.create_dataset("theta", data=self._theta, compression="lzf")
        f.create_dataset("phi", data=self._phi, compression="lzf")

        aux_grp = f.create_group("auxiliary_quantities")

        for k, v in self._auxiliary_quantities.items():

            q_grp = aux_grp.create_group(k)
            q_grp.create_dataset("true_values",
                                 data=v["true_values"],
                                 compression="lzf")
            q_grp.create_dataset("obs_values",
                                 data=v["obs_values"],
                                 compression="lzf")

        model_grp = f.create_group("model_spaces")

        for k, v in self._model_spaces.items():

            model_grp.create_dataset(k, data=v, compression="lzf")

        # now store the truths
        recursively_save_dict_contents_to_group(f, "truth", self._truth)

        recursively_save_dict_contents_to_group(
            f, "graph", fill_graph_dict(nx.to_dict_of_dicts(self._graph)))

        recursively_save_dict_contents_to_group(f, "popsynth", self._pop_synth)

        # now store the popsymnth

    @classmethod
    def from_file(cls, file_name):
        """
        load a population from a file
        :param file_name:
        :returns:
        :rtype:

        """

        with h5py.File(file_name, "r") as f:

            return cls._loadfrom(f)

    @classmethod
    def from_group(cls, file_name, group_name):
        """
        load a population from a group
        in a file.
        """

        with h5py.File(file_name, "r") as f:

            g = f[group_name]

            return cls._loadfrom(g)

    @classmethod
    def _loadfrom(cls, f):
        """
        load from file or group

        :param f: h5py file or group handle
        """

        spatial_params = {}

        for key in f["spatial_params"].keys():

            spatial_params[key] = f["spatial_params"][key][()][0]

        # we must double check that there are LF params

        try:

            if f.attrs["has_lf"]:
                lf_params = {}
                for key in f["lf_params"].keys():

                    lf_params[key] = f["lf_params"][key][()][0]
                lf_form = str(f.attrs["lf_form"])

            else:

                lf_params = None
                lf_form = None
        except:

            lf_params = None
            lf_form = None

        flux_sigma = f.attrs["flux_sigma"]
        n_model = f.attrs["n_model"]
        r_max = f.attrs["r_max"]
        seed = int(f.attrs["seed"])
        name = f.attrs["name"]

        spatial_form = str(f.attrs["spatial_form"])

        luminosities = f["luminosities"][()]
        distances = f["distances"][()]
        theta = f["theta"][()]
        phi = f["phi"][()]

        # right now this is just for older pops
        try:
            known_distances = f["known_distances"][()]
            known_distance_idx = (f["known_distance_idx"][()]).astype(int)
            unknown_distance_idx = (f["unknown_distance_idx"][()]).astype(int)

        except:

            known_distances = None
            known_distance_idx = None
            unknown_distance_idx = None

        fluxes = f["fluxes"][()]
        flux_obs = f["flux_obs"][()]
        selection = f["selection"][()]

        model_spaces = {}

        for k in f["model_spaces"].keys():

            model_spaces[str(k)] = f["model_spaces"][k][()]

        auxiliary_quantities = {}

        for k in f["auxiliary_quantities"].keys():

            auxiliary_quantities[str(k)] = {
                "true_values": f["auxiliary_quantities"][k]["true_values"][()],
                "obs_values": f["auxiliary_quantities"][k]["obs_values"][()],
            }

        truth = recursively_load_dict_contents_from_group(f, "truth")

        graph = nx.from_dict_of_dicts(
            clean_graph_dict(
                recursively_load_dict_contents_from_group(f, "graph")))

        pop_synth = recursively_load_dict_contents_from_group(f, "popsynth")

        return cls(
            luminosities=luminosities,
            distances=distances,
            known_distances=known_distances,
            known_distance_idx=known_distance_idx,
            unknown_distance_idx=unknown_distance_idx,
            fluxes=fluxes,
            flux_obs=flux_obs,
            selection=selection,
            flux_sigma=flux_sigma,
            n_model=n_model,
            r_max=r_max,
            lf_params=lf_params,
            spatial_params=spatial_params,
            model_spaces=model_spaces,
            seed=seed,
            name=name,
            spatial_form=spatial_form,
            lf_form=lf_form,
            auxiliary_quantities=auxiliary_quantities,
            truth=truth,
            graph=graph,
            theta=theta,
            phi=phi,
            pop_synth=pop_synth
        )

    def to_sub_population(self, observed: bool = True) -> "Population":
        """
        Create a population that is down selected from either the
        observed or unobserved population

        :param observed: extract the observed or unobserved object

        """

        if observed:
            selection = self._selection

        else:

            selection = ~self._selection

        if self._auxiliary_quantities is not None:

            new_aux = {}

            for k, v in self._auxiliary_quantities.items():

                new_aux[k] = {
                    "true_values": v["true_values"][selection],
                    "obs_values": v["obs_values"][selection],
                }

        else:

            new_aux = None

        itr = 0
        known_distances = []
        known_distance_idx = []
        unknown_distance_idx = []

        for i, s in enumerate(selection):

            if s:

                if i in self._known_distances:

                    known_distances.append(self._distances[i])
                    known_distance_idx.append(itr)

                else:

                    unknown_distance_idx.append(itr)

                itr += 1

        return Population(
            luminosities=self._luminosities[selection],
            distances=self._distances[selection],
            known_distances=np.array(known_distances),
            known_distance_idx=np.array(known_distance_idx),
            unknown_distance_idx=np.array(unknown_distance_idx),
            fluxes=self._fluxes[selection],
            flux_obs=self._flux_obs[selection],
            selection=np.ones(sum(selection), dtype=bool),
            flux_sigma=self._flux_sigma,
            n_model=self._n_model,
            r_max=self._r_max,
            lf_params=self._lf_params,
            spatial_params=self._spatial_params,
            model_spaces=self._model_spaces,
            seed=self._seed,
            name=self._name,
            spatial_form=self._spatial_form,
            lf_form=self._lf_form,
            auxiliary_quantities=new_aux,
            truth=self._truth,
            graph=self._graph,
            theta=self._theta[selection],
            phi=self._phi[selection],
        )

    def display(self):
        """
        Display the simulation parameters

        """

        info = "### %s simulation\nDetected %d out of %d objects" % (
            self._name,
            sum(self._selection),
            len(self._fluxes),
        )

        display(Markdown(info))

        if self._lf_params is not None:

            out = {"parameter": [], "value": []}

            display(Markdown("## Luminosity Function"))
            for k, v in self._lf_params.items():

                out["parameter"].append(k)
                out["value"].append(v)

            display(Math(self._lf_form))
            display(pd.DataFrame(out))
        out = {"parameter": [], "value": []}

        display(Markdown("## Spatial Function"))

        for k, v in self._spatial_params.items():

            out["parameter"].append(k)
            out["value"].append(v)

        display(Math(self._spatial_form))
        display(pd.DataFrame(out))

    def display_true_fluxes(self, ax=None, flux_color=dark, **kwargs):
        """Display the fluxes

        :param ax:
        :param flux_color:
        :returns:
        :rtype:

        """

        if ax is None:
            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()

        ax.scatter(
            self._distances,
            self._fluxes,
            alpha=0.5,
            color=flux_color,
            edgecolors="none",
            label="True flux",
            **kwargs,
        )

#        ax.axhline(self._boundary, color="grey", zorder=-5000, ls="--")

        # ax.set_xscale('log')
        ax.set_yscale("log")

        try:

            ax.set_ylim(
                bottom=min([self._fluxes.min(),
                            self._flux_selected.min()]))

        except:

            ax.set_ylim(bottom=self._fluxes.min())

        ax.set_xlabel("distance")
        ax.set_ylabel("flux")

    def display_obs_fluxes(self, ax=None, flux_color=dark, **kwargs):
        """FIXME! briefly describe function

        :param ax:
        :param flux_color:
        :returns:
        :rtype:

        """

        # do not try to plot if there is nothing
        # to plot

        if self._no_detection:
            log.warning("There are no detections to display")
            if ax is not None:

                fig = ax.get_figure()

                return fig
            else:

                return

        if ax is None:
            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()

        ax.scatter(
            self._distance_selected,
            self._flux_selected,
            alpha=0.8,
            color=flux_color,
            edgecolors="none",
            label="Detected flux",
            **kwargs,
        )

#        ax.axhline(self._boundary, color="grey", zorder=-5000, ls="--")
        # ax.set_xscale('log')
        ax.set_yscale("log")

        ax.set_ylim(bottom=min([self._fluxes.min(),
                                self._flux_selected.min()]))
        ax.set_xlim(right=self._r_max)

        ax.set_xlabel("distance")
        ax.set_ylabel("flux")
        return fig

    def display_fluxes(
        self,
        ax=None,
        true_color=light,
        obs_color=dark,
        arrow_color="k",
        with_arrows=True,
        **kwargs,
    ):
        """FIXME! briefly describe function

        :param ax:
        :param true_color:
        :param obs_color:
        :param arrow_color:
        :param with_arrows:
        :returns:
        :rtype:

        """

        if ax is None:
            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()

        self.display_true_fluxes(ax=ax, flux_color=true_color, **kwargs)

        if not self._no_detection:
            self.display_obs_fluxes(ax=ax, flux_color=obs_color, **kwargs)

        if (with_arrows) and (not self._no_detection):
            for start, stop, z in zip(
                    self._fluxes[self._selection],
                    self._flux_selected,
                    self._distance_selected,
            ):

                x = z
                y = start
                dx = 0
                dy = stop - start

                ax.arrow(
                    x,
                    y,
                    dx,
                    dy,
                    color=arrow_color,
                    head_width=0.05,
                    head_length=0.2 * np.abs(dy),
                    length_includes_head=True,
                )

        return fig

    def display_luminosities(
        self,
        ax=None,
        true_color=light,
        obs_color=dark,
        **kwargs,
    ):
        """FIXME! briefly describe function

        :param ax:
        :param true_color:
        :param obs_color:

        :returns:
        :rtype:

        """

        if ax is None:
            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()

        ax.scatter(self._distance_selected,
                   self._luminosity_selected,
                   s=5,
                   color=obs_color)
        ax.scatter(self._distance_hidden,
                   self._luminosity_hidden,
                   s=5,
                   color=true_color)

        return fig

    def _display_sphere(
        self,
        fluxes,
        distances,
        theta,
        phi,
        cmap="magma",
        distance_transform=None,
        use_log=False,
        fig=None,
        background_color="white",
        show=True,
        **kwargs,
    ):

        if len(fluxes) == 0:
            log.warning("There are no detections to display")

            return

        if fig is None:

            fig = ipv.figure()

        ipv.pylab.style.box_off()
        ipv.pylab.style.axes_off()
        ipv.pylab.style.set_style_dark()
        ipv.pylab.style.background_color(background_color)

        if distance_transform is not None:

            distance = distance_transform(distances)

        else:

            distance = distances

        x, y, z = _create_sphere_variables(self._r_max, distance, theta, phi)

        _, colors = array_to_cmap(fluxes, cmap, use_log=True)

        ipv.scatter(x, y, z, color=colors, marker="sphere", **kwargs)

        r_value = fig

        if show:

            ipv.xyzlim(self._r_max)
            fig.camera.up = [1, 0, 0]
            control = pythreejs.OrbitControls(controlling=fig.camera)
            fig.controls = control
            control.autoRotate = True
            fig.render_continuous = True
            #control.autoRotate = True
            # toggle_rotate = widgets.ToggleButton(description="Rotate")
            # widgets.jslink((control, "autoRotate"), (toggle_rotate, "value"))
            # r_value = toggle_rotate

        return fig

    def display_obs_fluxes_sphere(
        self,
        cmap="magma",
        distance_transform=None,
        use_log=False,
        background_color="white",
        show=True,
        **kwargs,
    ):

        theta = self._theta[self._selection]
        phi = self._phi[self._selection]

        fig = self._display_sphere(
            self._flux_selected,
            self._distance_selected,
            theta=theta,
            phi=phi,
            cmap=cmap,
            distance_transform=distance_transform,
            background_color=background_color,
            use_log=use_log,
            show=show,
            **kwargs,
        )

        if show:
            ipv.show()

        return fig

    def display_hidden_fluxes_sphere(
        self,
        cmap="magma",
        distance_transform=None,
        use_log=False,
        background_color="white",
        show=True,
        **kwargs,
    ):

        theta = self._theta[~self._selection]
        phi = self._phi[~self._selection]

        fig = self._display_sphere(
            self._flux_hidden,
            self._distance_hidden,
            theta=theta,
            phi=phi,
            cmap=cmap,
            distance_transform=distance_transform,
            background_color=background_color,
            use_log=use_log,
            show=show,
            **kwargs,
        )

        if show:
            ipv.show()

        return fig

    def display_flux_sphere(
        self,
        seen_cmap="Reds",
        unseen_cmap="Blues",
        distance_transform=None,
        use_log=False,
        background_color="white",
        **kwargs,
    ):

        fig = self.display_obs_fluxes_sphere(
            cmap=seen_cmap,
            distance_transform=distance_transform,
            use_log=use_log,
            background_color=background_color,
            show=False,
            **kwargs,
        )

        fig = self.display_hidden_fluxes_sphere(
            cmap=unseen_cmap,
            distance_transform=distance_transform,
            use_log=use_log,
            background_color=background_color,
            fig=fig,
            show=True,
            **kwargs,
        )

        return fig

    # def display_luminosty(self, ax=None, **kwargs):

    #     if self._no_detection:
    #         return

    #     if ax is None:
    #         fig, ax = plt.subplots()

    #     else:

    #         fig = ax.get_figure()

    #     bins = np.logspace(
    #         np.log10(self._luminosities.min()), np.log10(self._luminosities.max()), 30
    #     )

    #     ax.hist(
    #         self._luminosities,
    #         #bins=bins,
    #         #density=True,
    #         fc=dark,
    #         ec=dark_highlight,
    #         lw=1.5,
    #     )

    #     ax.set_xscale("log")
    #     ax.set_yscale("log")
    #     ax.set_xlabel("L")

    def display_distances(self, ax=None):
        """FIXME! briefly describe function

        :param ax:
        :returns:
        :rtype:

        """

        if ax is None:
            fig, ax = plt.subplots()

        else:
            fig = ax.get_figure()

        bins = np.linspace(0, self._r_max, 40)

        ax.hist(
            self._distances,
            bins=bins,
            fc=dark,
            ec=dark_highlight,
            lw=1.5,
            label="Total Pop.",
        )
        ax.hist(
            self._distance_selected,
            bins=bins,
            #            facecolor=blue,
            #            edgecolor=blue_highlight,
            lw=1.5,
            alpha=1,
            label="Obs. Pop.",
        )

        ax.set_xlabel("z")
        ax.legend()
        # sns.despine(offset=5, trim=True);


def _create_sphere_variables(R, distance, theta, phi):
    x, y, z = xyz(distance, theta, phi)

    # u = np.linspace(0, 2 * np.pi, 100)
    # v = np.linspace(0, np.pi, 100)
    # x2 = R * np.outer(np.cos(u), np.sin(v))
    # y2 = R * np.outer(np.sin(u), np.sin(v))
    # z2 = R * np.outer(np.ones(np.size(u)), np.cos(v))

    return x, y, z



from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from typing import Any, Dict, Optional, Union

import networkx as nx
import numpy as np
import pandas as pd
import scipy.integrate as integrate
import scipy.special as sf
import scipy.stats as stats
import yaml
from IPython.display import Markdown, Math, display
# from numpy.typing import np.ndarray
from numba import float64, jit, njit, prange

from popsynth.auxiliary_sampler import AuxiliarySampler, DerivedLumAuxSampler
from popsynth.distribution import LuminosityDistribution, SpatialDistribution
from popsynth.population import Population
from popsynth.selection_probability import (BernoulliSelection,
                                            HardFluxSelection,
                                            SelectionProbabilty,
                                            SoftFluxSelection, UnitySelection)
from popsynth.utils.logging import setup_logger
from popsynth.utils.progress_bar import progress_bar
from popsynth.utils.registry import (auxiliary_parameter_registry,
                                     distribution_registry, selection_registry)

log = setup_logger(__name__)


class PopulationSynth(object, metaclass=ABCMeta):
    def __init__(
        self,
        spatial_distribution: SpatialDistribution,
        luminosity_distribution: Optional[LuminosityDistribution] = None,
        seed: int = 1234,
    ):
        """
        Basic and generic population synth. One specifies the spatial and luminosity distribution OR
        derived luminosity distribution and everything is setup.

        :param spatial_distribution:
        :param luminosity_distribution:
        :param seed:
        :returns:
        :rtype:

        """

        self._n_model = 500  # type: int
        self._seed = int(seed)  # type: int

        # this is a container for things passed to
        # stan
        self._model_spaces = {}  # type: dict

        self._auxiliary_observations: Dict[str, AuxiliarySampler] = {}

        self._graph = nx.DiGraph()  # type: nx.Digraph

        self._name = f"{spatial_distribution.name}"  # type: str
        if luminosity_distribution is not None:
            self._name = f"{self._name}_{luminosity_distribution.name}"

            if not isinstance(
                luminosity_distribution, LuminosityDistribution
            ):

                log.error("the luminosity_distribution is the wrong type")

                raise RuntimeError()

        if not isinstance(
                spatial_distribution,
                SpatialDistribution):

            log.error("the spatial_distribution is the wrong type")

            raise RuntimeError()

        self._spatial_distribution = spatial_distribution  # type: SpatialDistribution
        self._luminosity_distribution = (
            luminosity_distribution
        )  # type: Union[LuminosityDistribution, None]

        self._has_derived_luminosity = False  # type: bool
        self._derived_luminosity_sampler = (
            None)  # type: Union[DerivedLumAuxSampler, None]

        # set the selections be fully seen unless it is set by the user
        self._distance_selector: SelectionProbabilty = UnitySelection(
            name="unity distance selector")
        self._flux_selector: SelectionProbabilty = UnitySelection(
            name="unity flux selector")

        # check to see if the selectors are set
        self._distance_selector_set: bool = False
        self._flux_selector_set: bool = False
        self._spatial_selector: Optional[SelectionProbabilty] = None

        self._params = {}  # type: dict

        # keep a list of parameters here for checking

        for k, v in self._spatial_distribution.params.items():

            self._params[k] = v

        if self._luminosity_distribution is not None:
            for k, v in self._luminosity_distribution.params.items():

                self._params[k] = v

        self._graph.add_node(self._spatial_distribution.name)

        # add the sky sampler

    def clean(self, reset: bool = False):
        """
        clean the auxiliary samplers, selections, etc 
        from the population synth
        :param reset: reset any attached distributions and samplers

        """

        if reset:

            for k, v in self._auxiliary_observations.items():

                v.reset()

        log.warning("removing all registered Auxiliary Samplers")

        self._auxiliary_observations = {}

        if reset:

            if self._derived_luminosity_sampler is not None:

                self._derived_luminosity_sampler.reset()

        self._derived_luminosity_sampler = None

        self._has_derived_luminosity = False

        self._distance_selector_set = False

        self._flux_selector_set = False

        log.warning("removing flux selector")

        if reset:

            self._flux_selector.reset()

        self._flux_selector = UnitySelection(name="unity flux selector")

        log.warning("removing distance selector")

        if reset:

            self._distance_selector.reset()

        self._distance_selector = UnitySelection(
            name="unity distance selector")

        log.warning("removing spatial selector")

        if reset:

            if self._spatial_selector is not None:

                self._spatial_selector.reset()

        self._spatial_selector = None

    def write_to(self, file_name: str) -> None:
        """
        write the population synth to a file
        :param file_name: the file name of the output yaml

        """
        with open(file_name, "w") as f:

            yaml.dump(stream=f, data=self.to_dict(),
                      default_flow_style=False, Dumper=yaml.SafeDumper)

    def to_dict(self) -> Dict[str, Any]:
        """
        convert the population synth to a dictionary
        """

        output: Dict[str, Any] = {}

        output["seed"] = self._seed

        # store the spatial distribution

        spatial_distribution = {}

        spatial_distribution[self._spatial_distribution._distribution_name] = self._spatial_distribution.truth

        output["spatial distribution"] = spatial_distribution

        # if there is a luminosity distribution
        # then get it and store it

        if self._luminosity_distribution is not None:

            luminosity_distribution = {}

            luminosity_distribution[self._luminosity_distribution._distribution_name] = self._luminosity_distribution.truth

            output["luminosity distribution"] = luminosity_distribution

        if self._flux_selector_set:

            flux_selection = {}

            flux_selection[self._flux_selector._selection_name] = self._flux_selector.parameters

            output["flux selection"] = flux_selection

        if self._distance_selector_set:

            distance_selection = {}

            distance_selection[self._distance_selector._selection_name] = self._distance_selector.parameters

            output["distance selection"] = distance_selection

        if self._spatial_selector is not None:

            spatial_selection = {}

            spatial_selection[self._spatial_selector._selection_name] = self._spatial_selector.parameters

            output["spatial selection"] = spatial_selection

        aux_samplers = {}

        for k, v in self._auxiliary_observations.items():

            tmp = {}

            tmp["name"] = v.name
            tmp["observed"] = v.observed

            for k2, v2 in v.truth.items():

                tmp[k2] = v2

            tmp["secondary"] = list(v.secondary_samplers.keys())

            selection = {}
            selection[v.selector._selection_name] = v.selector.parameters

            tmp["selection"] = selection

            aux_samplers[v._auxiliary_sampler_name] = tmp

        output["auxiliary samplers"] = aux_samplers

        return output

    @classmethod
    def from_dict(cls, input: Dict[str, Any]) -> "PopulationSynth":
        """
        build a PopulationSynth object from a dictionary

        :param input: the dictionary from which to build
        """

        if "luminosity distribution" in input:

            # try:

            # load the name and parameters

            tmp = input["luminosity distribution"]

            ld_name = list(tmp.keys())[0]

            # create the instance

            luminosity_distribtuion: LuminosityDistribution = distribution_registry[ld_name]

            # now set the values of the parameters

            log.debug(f"setting parameters for {ld_name}")

            for k, v in tmp[ld_name].items():

                log.debug(f"trying to set {k} to {v}")

                for x in luminosity_distribtuion.__class__.mro():

                    if k in x.__dict__:

                        setattr(luminosity_distribtuion, k, float(v))

                        break

                log.debug(f"{luminosity_distribtuion.params}")

        else:

            luminosity_distribtuion = None

        # try

        tmp = input["spatial distribution"]

        sd_name = list(tmp.keys())[0]

        spatial_distribtuion: SpatialDistribution = distribution_registry[sd_name]

        for k, v in tmp[sd_name].items():

            log.debug(f"trying to set {k} to {v}")

            for x in spatial_distribtuion.__class__.mro():

                if k in x.__dict__:

                    setattr(spatial_distribtuion, k, float(v))

                    break

        seed: int = input["seed"]

        # create the poopulation synth

        pop_synth: PopulationSynth = cls(
            spatial_distribtuion, luminosity_distribution=luminosity_distribtuion, seed=seed)

        # if there is a flux selection
        # then add it on

        if "flux selection" in input:

            tmp = input["flux selection"]

            if tmp is not None:

                fs_name = list(tmp.keys())[0]

                # extract the parameters

                params = tmp[fs_name]

                if params is None:

                    params = {}

                log.debug(f"flux selection parameters {params}")

                # make sure they are all floats

                fs = selection_registry.get(fs_name)

                for k, v in params.items():

                    log.debug(f"trying to set {k} to {v}")

                    for x in fs.__class__.mro():

                        if k in x.__dict__:

                            setattr(fs, k, float(v))

                            break

                # oh yes we are doing that
                fs._use_flux = True
                
                pop_synth.set_flux_selection(fs)

        if "distance selection" in input:

            tmp = input["distance selection"]

            if tmp is not None:

                ds_name = list(tmp.keys())[0]

                log.debug(f"adding distance selection {ds_name}")

                # extract the parameters

                params = tmp[ds_name]

                if params is None:

                    params = {}

                log.debug(f"distance selection parameters {params}")

                # make sure they are all floats

                ds = selection_registry.get(ds_name)

                for k, v in params.items():

                    log.debug(f"trying to set {k} to {v}")

                    for x in ds.__class__.mro():

                        if k in x.__dict__:

                            setattr(ds, k, float(v))

                            break

                ds._use_distance = True
                        
                pop_synth.set_distance_selection(ds)

        if "spatial selection" in input:

            tmp = input["spatial selection"]

            if tmp is not None:

                ss_name = list(tmp.keys())[0]

                log.debug(f"adding distance selection {ss_name}")

                # extract the parameters

                params = tmp[ss_name]

                if params is None:

                    params = {}

                log.debug(f"spatial selection parameters {params}")

                ss = selection_registry.get(ss_name)

                # make sure they are all floats

                for k, v in params.items():

                    log.debug(f"trying to set {k} to {v}")

                    for x in ss.__class__.mro():

                        if k in x.__dict__:

                            setattr(ss, k, float(v))

                            break

                pop_synth.add_spatial_selector(ss)

        # Now collect the auxiliary samplers

        # we will gather them so that we can sort out dependencies
        aux_samplers: Dict[str, AuxiliarySampler] = OrderedDict()
        secondary_samplers: [str, str] = OrderedDict()

        if "auxiliary samplers" in input:

            for k, v in input["auxiliary samplers"].items():

                # first we extract the required info

                sampler_name = k
                obj_name = v.pop("name")
                is_observed = v.pop("observed")

                # now we extract the selection
                # and secondary if they are there

                if "selection" in v:
                    selection = v.pop("selection")

                else:

                    selection = None

                if "secondary" in v:
                    secondary = list(np.atleast_1d(v.pop("secondary")))

                else:

                    secondary = None

                if "init variables" in v:
                    init_variables = v.pop("init variables")

                    if init_variables is None:

                        init_variables = {}

                else:

                    init_variables = {}

                # since we have popped everything
                # all that is left should be parameters

                params = v

                # now build the object

                try:

                    tmp: AuxiliarySampler = auxiliary_parameter_registry.get(
                        sampler_name,
                        name=obj_name,
                        observed=is_observed,

                        **init_variables)

                except(TypeError):

                    # try without name

                    try:

                        tmp: AuxiliarySampler = auxiliary_parameter_registry.get(
                            sampler_name,
                            # name=obj_name,
                            observed=is_observed,

                            **init_variables)

                    except(TypeError):

                        try:

                            tmp: AuxiliarySampler = auxiliary_parameter_registry.get(
                                sampler_name,
                                name=obj_name,
                                # observed=is_observed,

                                **init_variables)

                        except(TypeError):

                            tmp: AuxiliarySampler = auxiliary_parameter_registry.get(
                                sampler_name, **init_variables)

                            # we will do a dirty trick
                            tmp._name = obj_name
                            tmp._is_observed = is_observed

                log.debug(f"setting parameters for {sampler_name}: {obj_name}")

                # now set the parameters

                for k, v in params.items():

                    log.debug(f"trying to set {k} to {v}")

                    for x in tmp.__class__.mro():

                        if k in x.__dict__:

                            setattr(tmp, k, float(v))

                            break

                    log.debug(f"{tmp.truth}")

                if selection is not None:

                    sel_name = list(selection.keys())[0]

                    log.debug(f"adding selection {sel_name}")

                    # extract the parameters

                    params = selection[sel_name]

                    if params is None:

                        params = {}

                    log.debug(f"selection parameters {params}")

                    # make sure they are all floats

                    selector = selection_registry.get(sel_name)

                    for k, v in params.items():

                        if k in selector.__class__.__dict__:

                            setattr(selector, k, float(v))

                    selector._use_obs_value = True
                    
                    tmp.set_selection_probability(selector)

                # now we store this sampler

                log.debug(f"{obj_name} built")

                aux_samplers[obj_name] = tmp

                # if there is a secondary sampler,
                # we need to make a mapping

                if secondary is not None:

                    secondary_samplers[obj_name] = secondary

        # Now we have collected all of the auxiliary samplers
        # we need to assign those which are secondary

        log.debug(f"have {list(aux_samplers.keys())} as aux samplers")

        for primary, secondaries in secondary_samplers.items():

            for secondary in secondaries:

                if secondary is None:

                    break

                # assign it
                aux_samplers[primary].set_secondary_sampler(
                    aux_samplers[secondary])

        # now we need to pop all the secondaries from the main
        # list so that we do not double add

        for secondaries in list(secondary_samplers.values()):

            for secondary in secondaries:

                if secondary in aux_samplers:

                    aux_samplers.pop(secondary)

        # now we add the observed quantites on

        for k, v in aux_samplers.items():

            pop_synth.add_observed_quantity(v)

        return pop_synth

    @classmethod
    def from_file(cls, file_name: str) -> "PopulationSynth":
        """
        read the population in from a yaml file

        :param file_name: the file name of the population synth
        """
        with open(file_name) as f:

            input: Dict[str, Any] = yaml.load(f, Loader=yaml.SafeLoader)

        return cls.from_dict(input)

    @property
    def spatial_distribution(self) -> SpatialDistribution:
        return self._spatial_distribution

    @property
    def luminosity_distribution(self) -> Union[LuminosityDistribution, None]:
        return self._luminosity_distribution

    def add_model_space(self, name, start, stop, log=True):
        """
        Add a model space for stan generated quantities

        :param name: name that Stan will use
        :param start: start of the grid
        :param stop: stop of the grid
        :param log: use log10 or not

        """
        if log:
            space = np.logspace(np.log10(start), np.log10(stop), self._n_model)

        else:

            space = np.linspace(start, stop, self._n_model)

        self._model_spaces[name] = space

    def add_auxiliary_sampler(self,
                              auxiliary_sampler: Union[DerivedLumAuxSampler,
                                                       AuxiliarySampler]):
        """
        add an auxiliary sampler or derived luminosity sampler to the population
        synth

        :param auxiliary_sampler:
        :returns:
        :rtype:
        """

        self.add_observed_quantity(auxiliary_sampler)

    def add_observed_quantity(self,
                              auxiliary_sampler: Union[DerivedLumAuxSampler,
                                                       AuxiliarySampler]):
        """
        add an auxiliary sampler or derived luminosity sampler to the population
        synth

        :param auxiliary_sampler:
        :returns:
        :rtype:

        """

        if isinstance(auxiliary_sampler, DerivedLumAuxSampler):

            log.info(
                f"registering derived luminosity sampler: {auxiliary_sampler.name}")

            self._has_derived_luminosity = True
            self._derived_luminosity_sampler = auxiliary_sampler

        else:

            if auxiliary_sampler.is_secondary:
                log.error(
                    f"{auxiliary_sampler.name} is already set as a secondary sampler!")
                log.error(
                    f"and registered to {','.join(auxiliary_sampler.parents)}")

                raise RuntimeError()

            if auxiliary_sampler.name in self._auxiliary_observations:

                log.error(f"{auxiliary_sampler.name} is already registered!")

                raise RuntimeError()

            log.info("registering auxilary sampler: %s" %
                     auxiliary_sampler.name)

            self._auxiliary_observations[
                auxiliary_sampler.name] = auxiliary_sampler

    def set_distance_selection(self, selector: SelectionProbabilty) -> None:
        """
        Set the selection type for the distance
        """

        if not isinstance(selector, SelectionProbabilty):

            log.error(f"{selector} is not a Selection probability")

            raise RuntimeError()

        self._distance_selector = selector

        self._distance_selector_set = True

    def set_flux_selection(self, selector: SelectionProbabilty) -> None:
        """
        Set the selection type for the distance
        """
        if not isinstance(selector, SelectionProbabilty):

            log.error(f"{selector} is not a Selection probability")

            raise RuntimeError()

        self._flux_selector = selector

        self._flux_selector_set = True

    def add_spatial_selector(self,
                             spatial_selector: SelectionProbabilty) -> None:
        """
        Add a spatial selector into the mix
        """

        if not isinstance(spatial_selector, SelectionProbabilty):

            log.error(f"{spatial_selector} is not a Selection probability")

            raise RuntimeError()

        self._spatial_selector = spatial_selector

    @property
    def name(self) -> str:
        return self._name

    def draw_log10_fobs(self, f, f_sigma, size=1) -> np.ndarray:
        """
        draw the log10 of the the fluxes
        """

        log10_f = np.log10(f)

        # sample from the log distribution to keep positive fluxes
        log10_fobs = log10_f + np.random.normal(
            loc=0, scale=f_sigma, size=size)

        return log10_fobs

    def draw_log_fobs(self, f, f_sigma, size=1) -> np.ndarray:
        """
        draw the log10 of the the fluxes
        """

        log_f = np.log(f)

        # sample from the log distribution to keep positive fluxes
        log_fobs = log_f + np.random.normal(loc=0, scale=f_sigma, size=size)

        return log_fobs

    def draw_survey(
        self,
        flux_sigma: Optional[float] = None,
        log10_flux_draw: bool = True,
    ) -> Population:
        """
        Draw the total survey and return a Population object
        :param flux_sigma: the homoskedastic sigma for the flux in log10 space
        :return: a Population object
        """

        # this stores all the "true" population values from all the samplers
        truth = dict()  # type: dict

        # store the spatial distributVion truths
        truth[
            self._spatial_distribution.name] = self._spatial_distribution.truth

        # set the random seed

        np.random.seed(self._seed)

        # create a callback of the integrand
        dNdr = (lambda r: self._spatial_distribution.dNdV(
            r) * self._spatial_distribution.differential_volume(r) / self.
            _spatial_distribution.time_adjustment(r))

        # integrate the population to determine the true number of
        # objects
        N = integrate.quad(dNdr, 0.0,
                           self._spatial_distribution.r_max)[0]  # type: float

        log.info("The volume integral is %f" % N)

        # this should be poisson distributed
        n = np.random.poisson(N)  # type: np.int64

        self._spatial_distribution.draw_distance(size=n)

        # now draw the sky positions

        self._spatial_distribution.draw_sky_positions(size=n)

        distances = self._spatial_distribution.distances  # type: np.ndarray

        log.info("Expecting %d total objects" % n)

        # first check if the auxilliary samplers
        # compute the luminosities

        # setup the global selection

        global_selection: SelectionProbabilty = UnitySelection(name="global")
        global_selection.select(n)

        # now we set up the selection that _may_ come
        # from the auxilliary samplers

        auxiliary_selection: SelectionProbabilty = UnitySelection(
            name="total auxiliary selection")
        auxiliary_selection.select(n)

        auxiliary_quantities: Dict[str, Dict] = {}

        # this means the luminosity is not
        # simulated directy

        if self.luminosity_distribution is None:

            if not self._has_derived_luminosity:

                log.error("No luminosity distribution was specified")
                log.error(
                    "and no derived luminosity auxiliary sampler was added")
                raise RuntimeError()

        if self._has_derived_luminosity:

            log.debug("using a derived luminosity sampler")
            # pbar.set_description(desc='Getting derived luminosities')
            # set the distance to the auxilary sampler
            self._derived_luminosity_sampler.set_spatial_values(
                self._spatial_distribution.spatial_values)

            # sample the true and obs
            # values which are held internally

            self._derived_luminosity_sampler.draw(size=n)

            log.debug("derived luminosity sampled")

            # check to make sure we sampled!
            assert (self._derived_luminosity_sampler.true_values is not None
                    and len(self._derived_luminosity_sampler.true_values) == n)

            assert (self._derived_luminosity_sampler.obs_values is not None
                    and len(self._derived_luminosity_sampler.obs_values) == n)

            # append these values to a dict
            auxiliary_quantities[self._derived_luminosity_sampler.name] = {
                "true_values": self._derived_luminosity_sampler.true_values,
                "obs_values": self._derived_luminosity_sampler.obs_values,
                "selection": self._derived_luminosity_sampler.selector,
            }

            log.info("Getting luminosity from derived sampler")

            luminosities = (
                self._derived_luminosity_sampler.compute_luminosity()
            )  # type: np.ndarray

            # collect anything that was sampled here

            # store the truth from the derived lum sampler

            truth[self._derived_luminosity_sampler.
                  name] = self._derived_luminosity_sampler.truth

            log.debug("sampling ")

            for k2, v2 in self._derived_luminosity_sampler.secondary_samplers.items(
            ):

                # first we tell the sampler to go and retrieve all of
                # its own secondaries

                properties = v2.get_secondary_properties()  # type: dict

                for k3, v3 in properties.items():

                    # now attach them
                    auxiliary_quantities[k3] = v3

                # store the secondary truths
                # this will _could_ be clobbered later
                # but that is ok

                truth[v2.name] = v2.truth

        else:

            luminosities = self.luminosity_distribution.draw_luminosity(
                size=n)  # type: np.ndarray

            # store the truths from the luminosity distribution
            truth[self.luminosity_distribution.
                  name] = self.luminosity_distribution.truth

        # transform the fluxes
        fluxes = self._spatial_distribution.transform(
            luminosities, distances)  # type: np.ndarray

        # now sample any auxilary quantities
        # if needed

        for k, v in self._auxiliary_observations.items():

            assert (
                not v.is_secondary
            ), "This is a secondary sampler. You cannot sample it in the main sampler"

            # set the luminosities and distances to
            # auxilary sampler just in case
            # they are needed

            v.set_luminosity(luminosities)
            v.set_spatial_values(self._spatial_distribution.spatial_values)

            # sample the true and obs
            # values which are held internally
            # this will also invoke secondary samplers

            v.draw(size=n)

            # store the auxilliary truths
            truth[v.name] = v.truth

            # check to make sure we sampled!
            assert v.true_values is not None and len(v.true_values) == n
            assert v.obs_values is not None and len(v.obs_values) == n

            # append these values to a dict
            auxiliary_quantities[k] = {
                "true_values": v.true_values,
                "obs_values": v.obs_values,
                "selection": v.selector,
            }  # type: dict

            # collect the secondary values

            for k2, v2 in v.secondary_samplers.items():

                # first we tell the sampler to go and retrieve all of
                # its own secondaries

                properties = v2.get_secondary_properties()  # type: dict

                for k3, v3 in properties.items():

                    # now attach them
                    auxiliary_quantities[k3] = v3

                # store the secondary truths

                truth[v2.name] = v2.truth

        # pbar.update()

        # now draw all the observed fluxes
        # this is homoskedastic for now

        if not isinstance(self._flux_selector, UnitySelection):

            if flux_sigma is not None:

                log.debug("assuming that fluxes will be jittered")

                if log10_flux_draw:

                    log.debug("making a log10 flux draw")

                    log10_fluxes_obs = self.draw_log10_fobs(fluxes, flux_sigma,
                                                            size=n)  # type: np.ndarray
                    flux_obs = np.power(10, log10_fluxes_obs)

                else:

                    log.debug("making a logflux draw")

                    log10_fluxes_obs = self.draw_log_fobs(fluxes, flux_sigma,
                                                          size=n)  # type: np.ndarray

                    flux_obs = np.exp(log10_fluxes_obs)

                assert np.alltrue(np.isfinite(log10_fluxes_obs))

            else:

                log.debug("observed fluxes are latent fluxes")

                flux_obs = fluxes
                log10_fluxes_obs = np.log10(fluxes)
                flux_sigma = -1  # this is a dummy

        else:

            log.debug("observed fluxes are latent fluxes")

            flux_obs = fluxes
            log10_fluxes_obs = np.log10(fluxes)
            flux_sigma = -1  # this is a dummy

        log.info("applying selection to fluxes")

        # pass the values the plux selector and draw the selection
        self._flux_selector.set_observed_flux(flux_obs)

        self._flux_selector.select(n)

        #       selection = self._flux_selector.selection

        # now apply the selection from the auxilary samplers

        for k, v in auxiliary_quantities.items():

            auxiliary_selection += v["selection"]

            log.info(
                "Applying selection from %s which selected %d of %d objects" %
                (k, v["selection"].n_selected, v["selection"].n_objects))

            log.info(
                "Before auxiliary selection there were %d objects selected" %
                self._flux_selector.n_selected)

        # now we can add the values onto the global
        # selection
        # not in the future we will depreciate the
        # no selection feature

        global_selection += auxiliary_selection

        global_selection += self._flux_selector

        # now scan the spatial selector

        if (self._spatial_selector is not None):

            self._spatial_selector.set_spatial_distribution(
                self._spatial_distribution)

            self._spatial_selector.select(n)

            log.info(
                f"Appling selection from {self._spatial_selector.name} which selected {self._spatial_selector.n_selected} of {self._spatial_selector.n_objects}"
            )

            global_selection += self._spatial_selector

        if global_selection.n_selected == n:

            log.warning("NO HIDDEN OBJECTS")

        self._distance_selector.select(size=global_selection.n_selected)

        known_distances = distances[global_selection.selection][
            self._distance_selector.selection]
        known_distance_idx = self._distance_selector.selection_index
        unknown_distance_idx = self._distance_selector.non_selection_index

        log.info(f"Detected {len(known_distances)} distances")

        try:

            log.info(
                f"Detected {global_selection.n_selected} objects out to a distance of {max(known_distances):.2f}")

        except:

            log.warning("No Objects detected")

        # just to make sure we do not do anything nutty
        lf_params = None
        lf_form = None
        if self._luminosity_distribution is not None:

            lf_params = self._luminosity_distribution.params
            lf_form = self._luminosity_distribution.form

        # if distance_probability is None:
        #     distance_probability = 1.0

        return Population(
            luminosities=luminosities,
            distances=distances,
            known_distances=known_distances,
            known_distance_idx=known_distance_idx,
            unknown_distance_idx=unknown_distance_idx,
            fluxes=fluxes,
            flux_obs=flux_obs,
            selection=global_selection.selection,
            flux_sigma=flux_sigma,
            r_max=self._spatial_distribution.r_max,
            n_model=self._n_model,
            lf_params=lf_params,
            spatial_params=self._spatial_distribution.params,
            model_spaces=self._model_spaces,
            seed=self._seed,
            name=self._name,
            spatial_form=self._spatial_distribution.form,
            lf_form=lf_form,
            auxiliary_quantities=auxiliary_quantities,
            truth=truth,
            graph=self.graph,
            theta=self._spatial_distribution.theta,
            phi=self._spatial_distribution.phi,
            pop_synth=self.to_dict()
        )

    def display(self) -> None:
        """
        Display the simulation parameters

        """

        if self._luminosity_distribution is not None:

            out = {"parameter": [], "value": []}

            display(Markdown("## Luminosity Function"))
            for k, v in self._luminosity_distribution.params.items():

                out["parameter"].append(k)
                out["value"].append(v)

            display(Math(self._luminosity_distribution.form))
            display(pd.DataFrame(out))
        out = {"parameter": [], "value": []}

        display(Markdown("## Spatial Function"))

        for k, v in self._spatial_distribution.params.items():

            out["parameter"].append(k)
            out["value"].append(v)

        display(Math(self._spatial_distribution.form))
        display(pd.DataFrame(out))

    # def generate_stan_code(self, stan_gen, **kwargs):

    #     pass

    @property
    def graph(self):

        self._build_graph()

        return self._graph

    def _build_graph(self):
        """
        builds the graph for all the samplers

        :returns:
        :rtype:

        """

        # first check out the luminosity sampler

        self._graph.add_node("obs_flux", observed=True)
        self._graph.add_edge(self._spatial_distribution.name, "obs_flux")
        if self._has_derived_luminosity:

            self._graph.add_node(self._derived_luminosity_sampler.name)

            self._graph.add_edge(self._derived_luminosity_sampler.name,
                                 "obs_flux")

            if self._derived_luminosity_sampler.uses_distance:

                self._graph.add_edge(
                    self._spatial_distribution.name,
                    self._derived_luminosity_sampler.name,
                )

            for k2, v2 in self._derived_luminosity_sampler.secondary_samplers.items(
            ):

                self._graph.add_node(k2)
                self._graph.add_edge(k2, self._derived_luminosity_sampler.name)

                # pass the graph and the primary

                _ = v2.get_secondary_properties(
                    graph=self._graph,
                    primary=k2,
                    spatial_distribution=self._spatial_distribution,
                )

        else:
            self._graph.add_edge(self._luminosity_distribution.name,
                                 "obs_flux")
        # now do the same fro everything else

        for k, v in self._auxiliary_observations.items():

            assert (
                not v.is_secondary
            ), "This is a secondary sampler. You cannot sample it in the main sampler"

            self._graph.add_node(k, observed=False)

            if v.observed:
                self._graph.add_node(v.obs_name, observed=False)
                self._graph.add_edge(k, v.obs_name)

                if v.uses_distance:

                    self._graph.add_edge(self._spatial_distribution.name, k)

            for k2, v2 in v.secondary_samplers.items():

                # first we tell the sampler to go and retrieve all of
                # its own secondaries

                self._graph.add_edge(k2, k)
                self._graph.add_node(k2, observed=False)

                if v2.uses_distance:

                    self._graph.add_edge(self._spatial_distribution.name, k2)

                if v2.observed:

                    self._graph.add_node(v2.obs_name, observed=True)
                    self._graph.add_edge(k2, v2.obs_name)

                _ = v2.get_secondary_properties(
                    graph=self._graph,
                    primary=k2,
                    spatial_distribution=self._spatial_distribution,
                )



import numpy as np

from popsynth.auxiliary_sampler import NonObservedAuxSampler
from popsynth.utils.spherical_geometry import sample_theta_phi


class SkySampler(object):
    _auxiliary_sampler_name = "SkySampler"
    def __init__(self, ra_sampler=None, dec_sampler=None):
        """
        A sky sampler that samples theta phi positions

        :returns:
        :rtype:

        """

        self._ra_sampler = ra_sampler
        self._dec_sampler = dec_sampler

        self._setup_sky()

    def _setup_sky(self):

        if self._ra_sampler is None:
            self._ra_sampler = RASampler()

        if self._dec_sampler is None:
            self._dec_sampler = DecSampler()

    @property
    def ra_sampler(self):

        return self._ra_sampler

    @property
    def dec_sampler(self):

        return self._dec_sampler


class RASampler(NonObservedAuxSampler):
    _auxiliary_sampler_name = "RASampler"
    def __init__(self):
        """

        Samples the RA of sky uniformly

        :param true_values:
        :returns:
        :rtype:

        """

        super(RASampler, self).__init__(name="ra")

    def true_sampler(self, size):

        self._true_values = np.random.uniform(0, 2 * np.pi, size=size)


class DecSampler(NonObservedAuxSampler):
    _auxiliary_sampler_name = "DecSampler"
    def __init__(self):
        """
        sampler the dec of the sky uniformly


        :param true_values:
        :returns:
        :rtype:

        """

        super(DecSampler, self).__init__(name="dec")

    def true_sampler(self, size):

        self._true_values = np.arccos(1 - 2 *
                                      np.random.uniform(0.0, 1.0, size=size))



import abc
from typing import Any, Dict, Optional

import numpy as np
from class_registry import AutoRegister

from popsynth.utils.logging import setup_logger
from popsynth.utils.meta import Parameter, ParameterMeta
from popsynth.utils.registry import selection_registry

log = setup_logger(__name__)


class SelectionParameter(Parameter):
    pass


class SelectionProbabilty(object, metaclass=AutoRegister(selection_registry, base_type=ParameterMeta)):
    def __init__(self, name: str = "name", use_obs_value: bool = False, use_distance: bool = False, use_luminosity: bool = False, use_flux: bool = False) -> None:

        self._name = name  # type: str

        self._parameter_storage = {}

        self._observed_flux = None  # type: np.ndarray
        self._observed_value = None  # type: np.ndarray
        self._distance = None  # type: np.ndarray
        self._luminosity = None  # type: np.ndarray
        self._selection = None  # type: np.ndarray
        self._is_sampled: bool = False

        self._use_obs_value: bool = use_obs_value
        self._use_distance: bool = use_distance
        self._use_luminosity: bool = use_luminosity
        self._use_flux: bool = use_flux

    def __add__(self, other):

        log.debug(f"adding selection from {other.name} to {self.name}")

        new_selection = np.logical_and(self._selection, other.selection)

        out = DummySelection()
        out._selection = new_selection
        return out

    # def __radd__(self, other):

    #     return self.__add__(other)

    def set_luminosity(self, luminosity: np.ndarray) -> None:
        """FIXME! briefly describe function

        :param luminosity:
        :returns:
        :rtype:

        """

        self._luminosity = luminosity  # type: np.ndarray

    def set_distance(self, distance: np.ndarray) -> None:
        """FIXME! briefly describe function

        :param distance:
        :returns:
        :rtype:

        """

        self._distance = distance  # type: np.ndarray

    def set_observed_flux(self, observed_flux: np.ndarray) -> None:
        """FIXME! briefly describe function

        :param luminosity:
        :returns:
        :rtype:

        """

        self._observed_flux = observed_flux  # type: np.ndarray

    def set_observed_value(self, observed_value: np.ndarray) -> None:
        self._observed_value = observed_value  # type: np.ndarray

    @abc.abstractclassmethod
    def draw(self, size: int) -> None:

        pass

    def reset(self):
        """
        resest the selector
        """
        if self._is_sampled:

            log.info(f"Selection: {self.name} is being reset")

            self._is_sampled = False
            self._selection = None

    def select(self, size: int):

        if not self._is_sampled:

            self.draw(size)

        else:

            log.warning(f"selecting with {self.name} more than once!")

        log.debug(f"{self.name} sampled {sum(self._selection)} objects")

        self._is_sampled = True

    @property
    def parameters(self) -> Dict[str, float]:

        out = {}

        for k, v in self._parameter_storage.items():

            if v is not None:

                out[k] = v

        return out

    @property
    def selection(self) -> np.ndarray:
        if self._selection is not None:

            return self._selection

        else:

            log.error(f"selector: {self.name} as not be sampled yet!")
            raise RuntimeError()

    @property
    def n_selected(self) -> int:
        return sum(self._selection)

    @property
    def n_non_selected(self) -> int:
        return sum(~self._selection)

    @property
    def n_objects(self) -> int:
        return self._selection.shape[0]

    @property
    def selection_index(self) -> np.ndarray:
        return np.where(self._selection)[0]

    @property
    def non_selection_index(self) -> np.ndarray:
        return np.where(~self._selection)[0]

    @property
    def name(self) -> str:
        return self._name


class DummySelection(SelectionProbabilty):
    _selection_name = "DummySelection"

    def __init__(self):

        super(DummySelection, self).__init__(name="dummy")

    def draw(self, size=1):
        pass
